#  Copyright (c) 2023 zfit
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Root search functions."""

from typing import Callable

import numpy as np
import tensorflow as tf

from . import tff_types as types
from .tff_dataclasses import dataclass as tff_dataclass

__all__ = [
    "BrentResults",
    "brentq",
]


def default_relative_root_tolerance(dtype):
    """Returns the default relative root tolerance used for a TensorFlow dtype."""
    if dtype is None:
        dtype = tf.float64
    return 4 * np.finfo(dtype.as_numpy_dtype(0)).eps


# TODO(b/179451420): Refactor BrentResults as RootSearchResults and return it
# for newton method as well.
@tff_dataclass
class BrentResults:
    """Brent root search results.

    Attributes:
      estimated_root: A `Tensor` containing the best estimate. If the search was
        successful, this estimate is a root of the objective function.
      objective_at_estimated_root: A `Tensor` containing the value of the
        objective function at  the best estimate. If the search was successful,
        then this is close to 0.
      num_iterations: A `Tensor` containing number of iterations performed for
        each pair of starting points
      converged: A boolean `Tensor` indicating whether the best estimate is a root
        within the tolerance specified for the search.
    """

    estimated_root: types.RealTensor
    objective_at_estimated_root: types.RealTensor
    num_iterations: types.IntTensor
    converged: types.BoolTensor


@tff_dataclass
class _BrentSearchConstants:
    """Values which remain fixed across all root searches."""

    false: types.BoolTensor
    zero: types.RealTensor
    zero_value: types.RealTensor


@tff_dataclass
class _BrentSearchState:
    """Values which are updated during the root search."""

    best_estimate: types.RealTensor
    value_at_best_estimate: types.RealTensor
    last_estimate: types.RealTensor
    value_at_last_estimate: types.RealTensor
    contrapoint: types.RealTensor
    value_at_contrapoint: types.RealTensor
    step_to_best_estimate: types.RealTensor
    step_to_last_estimate: types.RealTensor
    num_iterations: types.IntTensor
    finished: types.BoolTensor


@tff_dataclass
class _BrentSearchParams:
    """Values which remain fixed for a given root search."""

    objective_fn: Callable[[types.BoolTensor], types.BoolTensor]
    max_iterations: types.IntTensor
    absolute_root_tolerance: types.RealTensor
    relative_root_tolerance: types.RealTensor
    function_tolerance: types.RealTensor
    stopping_policy_fn: Callable[[types.BoolTensor], types.BoolTensor]


def _swap_where(condition, x, y):
    """Swaps the elements of `x` and `y` based on `condition`.

    Args:
      condition: A `Tensor` of dtype bool.
      x: A `Tensor` with the same shape as `condition`.
      y: A `Tensor` with the same shape and dtype as `x`.

    Returns:
      Two `Tensors` with the same shape as `x` and `y`.
    """
    return tf.where(condition, y, x), tf.where(condition, x, y)


def _secant_step(x1, x2, y1, y2):
    """Returns the step size at the current position if using the secant method.

    This function is meant for exclusive use by the `_brent_loop_body` function:
    - It does not guard against divisions by zero, and instead assumes that `y1`
      is distinct from `y2`. The `_brent_loop_body` function guarantees this
      property.
    - It does not guard against overflows which may occur if the difference
      between `y1` and `y2` is small while that between `x1` and `x2` is not.
      In this case, the resulting step size will be larger than `bisection_step`
      and thus ignored by the `_brent_loop_body` function.

    Args:
      x1: `Tensor` containing the current position.
      x2: `Tensor` containing the previous position.
      y1: `Tensor` containing the value of `objective_fn` at `x1`.
      y2: `Tensor` containing the value of `objective_fn` at `x2`.

    Returns:
      A `Tensor` with the same shape and dtype as `current`.
    """
    x_difference = x1 - x2
    y_difference = y1 - y2
    return -y1 * x_difference / y_difference


def _quadratic_interpolation_step(x1, x2, x3, y1, y2, y3):
    """Returns the step size to use when using quadratic interpolation.

    This function is meant for exclusive use by the `_brent_loop_body` function.
    It does not guard against divisions by zero, and instead assumes that `y1` is
    distinct from `y2` and `y3`. The `_brent_loop_body` function guarantees this
    property.

    Args:
      x1: `Tensor` of any shape and real dtype containing the first position used
        for extrapolation.
      x2: `Tensor` of the same shape and dtype as `x1` containing the second
        position used for extrapolation.
      x3: `Tensor` of the same shape and dtype as `x1` containing the third
        position used for extrapolation.
      y1: `Tensor` containing the value of the interpolated function at `x1`.
      y2: `Tensor` containing the value of interpolated function at `x2`.
      y3: `Tensor` containing the value of interpolated function at `x3`.

    Returns:
      A `Tensor` with the same shape and dtype as `x1`.
    """
    r2 = (x2 - x1) / (y2 - y1)
    r3 = (x3 - x1) / (y3 - y1)
    return -x1 * tf.math.divide_no_nan(x3 * r3 - x2 * r2, r3 * r2 * (x3 - x2))


def _should_stop(state, stopping_policy_fn):
    """Indicates whether the overall Brent search should continue.

    Args:
      state: A Python `_BrentSearchState` namedtuple.
      stopping_policy_fn: Python `callable` controlling the algorithm termination.

    Returns:
      A boolean value indicating whether the overall search should continue.
    """
    return tf.convert_to_tensor(
        stopping_policy_fn(state.finished), name="should_stop", dtype=tf.bool
    )


# This is a direct translation of the Brent root-finding method.
# Each operation is guarded by a call to `tf.where` to avoid performing
# unnecessary calculations.
def _brent_loop_body(state, params, constants):
    """Performs one iteration of the Brent root-finding algorithm.

    Args:
      state: A Python `_BrentSearchState` namedtuple.
      params: A Python `_BrentSearchParams` namedtuple.
      constants: A Python `_BrentSearchConstants` namedtuple.

    Returns:
      The `Tensor`s to use for the next iteration of the algorithm.
    """

    best_estimate = state.best_estimate
    last_estimate = state.last_estimate
    contrapoint = state.contrapoint
    value_at_best_estimate = state.value_at_best_estimate
    value_at_last_estimate = state.value_at_last_estimate
    value_at_contrapoint = state.value_at_contrapoint
    step_to_best_estimate = state.step_to_best_estimate
    step_to_last_estimate = state.step_to_last_estimate
    num_iterations = state.num_iterations
    finished = state.finished

    # If the root is between the last two estimates, use the worst of the two
    # as new contrapoint. Adjust step sizes accordingly.
    replace_contrapoint = ~finished & (
        value_at_last_estimate * value_at_best_estimate < constants.zero_value
    )

    contrapoint = tf.where(replace_contrapoint, last_estimate, contrapoint)
    value_at_contrapoint = tf.where(
        replace_contrapoint, value_at_last_estimate, value_at_contrapoint
    )

    step_to_last_estimate = tf.where(
        replace_contrapoint, best_estimate - last_estimate, step_to_last_estimate
    )
    step_to_best_estimate = tf.where(
        replace_contrapoint, step_to_last_estimate, step_to_best_estimate
    )

    # If the contrapoint is a better guess than the current root estimate, swap
    # them. Also, replace the worst of the two with the current contrapoint.
    replace_best_estimate = tf.where(
        finished,
        constants.false,
        tf.math.abs(value_at_contrapoint) < tf.math.abs(value_at_best_estimate),
    )

    last_estimate = tf.where(replace_best_estimate, best_estimate, last_estimate)
    best_estimate = tf.where(replace_best_estimate, contrapoint, best_estimate)
    contrapoint = tf.where(replace_best_estimate, last_estimate, contrapoint)

    value_at_last_estimate = tf.where(
        replace_best_estimate, value_at_best_estimate, value_at_last_estimate
    )
    value_at_best_estimate = tf.where(
        replace_best_estimate, value_at_contrapoint, value_at_best_estimate
    )
    value_at_contrapoint = tf.where(
        replace_best_estimate, value_at_last_estimate, value_at_contrapoint
    )

    # Compute the tolerance used to control root search at the current position
    # and the step size corresponding to the bisection method.
    root_tolerance = 0.5 * (
        params.absolute_root_tolerance
        + params.relative_root_tolerance * tf.math.abs(best_estimate)
    )
    bisection_step = 0.5 * (contrapoint - best_estimate)

    # Mark the search as finished if either:
    # 1. the maximum number of iterations has been reached;
    # 2. the desired tolerance has been reached (even if no root was found);
    # 3. the current root estimate is good enough.
    # Using zero as `function_tolerance` will check for exact roots and match
    # both Brent's original algorithm and the SciPy implementation.
    finished |= (
        (num_iterations >= params.max_iterations)
        | (tf.math.abs(bisection_step) < root_tolerance)
        | (~tf.math.is_finite(value_at_best_estimate))
        | (tf.math.abs(value_at_best_estimate) <= params.function_tolerance)
    )

    # Determine whether interpolation or extrapolation are worth performing at
    # the current position.
    compute_short_step = tf.where(
        finished,
        constants.false,
        (root_tolerance < tf.math.abs(step_to_last_estimate))
        & (tf.math.abs(value_at_best_estimate) < tf.math.abs(value_at_last_estimate)),
    )

    short_step = tf.where(
        compute_short_step,
        tf.where(
            # The contrapoint cannot be equal to the current root estimate since
            # they have opposite signs. However, it may be equal to the previous
            # estimate.
            tf.equal(last_estimate, contrapoint),
            # If so, use the secant method to avoid a division by zero which
            # would occur if using extrapolation.
            _secant_step(
                best_estimate,
                last_estimate,
                value_at_best_estimate,
                value_at_last_estimate,
            ),
            # Pass values of the objective function as x values, and root
            # estimates as y values in order to perform *inverse* extrapolation.
            _quadratic_interpolation_step(
                value_at_best_estimate,
                value_at_last_estimate,
                value_at_contrapoint,
                best_estimate,
                last_estimate,
                contrapoint,
            ),
        ),
        # Default to zero if using bisection.
        constants.zero,
    )

    # Use the step calculated above if both:
    # 1. step size < |previous step size|
    # 2. step size < 3/4 * |contrapoint - current root estimate|
    # Ensure that `short_step` was calculated by guarding the calculation with
    # `compute_short_step`.
    use_short_step = tf.where(
        compute_short_step,
        2 * tf.math.abs(short_step)
        < tf.minimum(
            3 * tf.math.abs(bisection_step) - root_tolerance,
            tf.math.abs(step_to_last_estimate),
        ),
        constants.false,
    )

    # Revert to bisection when not using `short_step`.
    step_to_last_estimate = tf.where(
        use_short_step, step_to_best_estimate, bisection_step
    )
    step_to_best_estimate = tf.where(
        finished, constants.zero, tf.where(use_short_step, short_step, bisection_step)
    )

    # Update the previous and current root estimates.
    last_estimate = tf.where(finished, last_estimate, best_estimate)
    best_estimate += tf.where(
        finished,
        constants.zero,
        tf.where(
            root_tolerance < tf.math.abs(step_to_best_estimate),
            step_to_best_estimate,
            tf.where(bisection_step > 0, root_tolerance, -root_tolerance),
        ),
    )

    value_at_last_estimate = tf.where(
        finished, value_at_last_estimate, value_at_best_estimate
    )
    value_at_best_estimate = tf.where(
        finished, value_at_best_estimate, params.objective_fn(best_estimate)
    )

    num_iterations = tf.where(finished, num_iterations, num_iterations + 1)

    return [
        _BrentSearchState(
            best_estimate=best_estimate,
            last_estimate=last_estimate,
            contrapoint=contrapoint,
            value_at_best_estimate=value_at_best_estimate,
            value_at_last_estimate=value_at_last_estimate,
            value_at_contrapoint=value_at_contrapoint,
            step_to_best_estimate=step_to_best_estimate,
            step_to_last_estimate=step_to_last_estimate,
            num_iterations=num_iterations,
            finished=finished,
        )
    ]


def _prepare_brent_args(
    objective_fn,
    left_bracket,
    right_bracket,
    value_at_left_bracket,
    value_at_right_bracket,
    absolute_root_tolerance=2e-7,
    relative_root_tolerance=None,
    function_tolerance=2e-7,
    max_iterations=100,
    stopping_policy_fn=None,
):
    r"""Prepares arguments for root search using Brent's method.

    Args:
      objective_fn: Python callable for which roots are searched. It must be a
        callable of a single `Tensor` parameter and return a `Tensor` of the same
        shape and dtype as `left_bracket`.
      left_bracket: `Tensor` or Python float representing the first starting
        points. The function will search for roots between each pair of points
        defined by `left_bracket` and `right_bracket`. The shape of `left_bracket`
        should match that of the input to `objective_fn`.
      right_bracket: `Tensor` of the same shape and dtype as `left_bracket` or
        Python float representing the second starting points. The function will
        search for roots between each pair of points defined by `left_bracket` and
        `right_bracket`. This argument must have the same shape as `left_bracket`.
      value_at_left_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `left_bracket`. If specified, this argument
        must have the same shape as `left_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      value_at_right_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `right_bracket`. If specified, this argument
        must have the same shape as `right_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      absolute_root_tolerance: Optional `Tensor` representing the absolute
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `2e-7`.
      relative_root_tolerance: Optional `Tensor` representing the relative
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `None` which translates to `4 *
          numpy.finfo(left_bracket.dtype.as_numpy_dtype).eps`.
      function_tolerance: Optional `Tensor` representing the tolerance used to
        check for roots. If the absolute value of `objective_fn` is smaller than
        or equal to `function_tolerance` at a given estimate, then that estimate
        is considered a root for the function. If specified, this argument must
        broadcast with the shape of `left_bracket` and have the same dtype. Set to
        zero to match Brent's original algorithm and to continue the search until
        an exact root is found.
        Default value: `2e-7`.
      max_iterations: Optional `Tensor` of an integral dtype or Python integer
        specifying the maximum number of steps to perform for each initial point.
        Must broadcast with the shape of `left_bracket`. If an element is set to
        zero, the function will not search for any root for the corresponding
        points in `left_bracket` and `right_bracket`. Instead, it will return the
        best estimate from the inputs.
        Default value: `100`.
      stopping_policy_fn: Python `callable` controlling the algorithm termination.
        It must be a callable accepting a `Tensor` of booleans with the shape of
        `left_bracket` (each denoting whether the search is finished for each
        starting point), and returning a scalar boolean `Tensor` (indicating
        whether the overall search should stop). Typical values are
        `tf.reduce_all` (which returns only when the search is finished for all
        pairs of points), and `tf.reduce_any` (which returns as soon as the search
        is finished for any pair of points).
        Default value: `None` which translates to `tf.reduce_all`.

    Returns:
      A tuple of 3 Python objects containing the state, parameters, and constants
      to use for the search.
    """
    stopping_policy_fn = stopping_policy_fn or tf.reduce_all
    if not callable(stopping_policy_fn):
        raise ValueError("stopping_policy_fn must be callable")

    left_bracket = tf.convert_to_tensor(left_bracket, name="left_bracket")
    right_bracket = tf.convert_to_tensor(
        right_bracket, name="right_bracket", dtype=left_bracket.dtype
    )

    if value_at_left_bracket is None:
        value_at_left_bracket = objective_fn(left_bracket)
    if value_at_right_bracket is None:
        value_at_right_bracket = objective_fn(right_bracket)

    value_at_left_bracket = tf.convert_to_tensor(
        value_at_left_bracket,
        name="value_at_left_bracket",
        dtype=left_bracket.dtype.base_dtype,
    )
    value_at_right_bracket = tf.convert_to_tensor(
        value_at_right_bracket,
        name="value_at_right_bracket",
        dtype=left_bracket.dtype.base_dtype,
    )

    if relative_root_tolerance is None:
        relative_root_tolerance = default_relative_root_tolerance(
            left_bracket.dtype.base_dtype
        )

    absolute_root_tolerance = tf.convert_to_tensor(
        absolute_root_tolerance,
        name="absolute_root_tolerance",
        dtype=left_bracket.dtype,
    )
    relative_root_tolerance = tf.convert_to_tensor(
        relative_root_tolerance,
        name="relative_root_tolerance",
        dtype=left_bracket.dtype,
    )
    function_tolerance = tf.convert_to_tensor(
        function_tolerance, name="function_tolerance", dtype=left_bracket.dtype
    )

    max_iterations = tf.broadcast_to(
        tf.convert_to_tensor(max_iterations),
        name="max_iterations",
        shape=left_bracket.shape,
    )
    num_iterations = tf.zeros_like(max_iterations)

    false = tf.constant(False, shape=left_bracket.shape)

    zero = tf.zeros_like(left_bracket)
    contrapoint = zero
    step_to_last_estimate = zero
    step_to_best_estimate = zero

    zero_value = tf.zeros_like(value_at_left_bracket)
    value_at_contrapoint = zero_value

    # Select the best root estimates from the inputs.
    # If no search is performed (e.g. `max_iterations` is `zero`), the estimate
    # computed this way will be returned. This differs slightly from the SciPy
    # implementation which always returns the `right_bracket`.
    swap_positions = tf.math.abs(value_at_left_bracket) < tf.math.abs(
        value_at_right_bracket
    )
    best_estimate, last_estimate = _swap_where(
        swap_positions, right_bracket, left_bracket
    )
    value_at_best_estimate, value_at_last_estimate = _swap_where(
        swap_positions, value_at_right_bracket, value_at_left_bracket
    )

    # Check if the current root estimate is good enough.
    # Using zero as `function_tolerance` will check for exact roots and match both
    # Brent's original algorithm and the SciPy implementation.
    finished = (
        (num_iterations >= max_iterations)
        | (~tf.math.is_finite(value_at_last_estimate))
        | (~tf.math.is_finite(value_at_best_estimate))
        | (tf.math.abs(value_at_best_estimate) <= function_tolerance)
    )

    return (
        _BrentSearchState(
            best_estimate=best_estimate,
            last_estimate=last_estimate,
            contrapoint=contrapoint,
            value_at_best_estimate=value_at_best_estimate,
            value_at_last_estimate=value_at_last_estimate,
            value_at_contrapoint=value_at_contrapoint,
            step_to_best_estimate=step_to_best_estimate,
            step_to_last_estimate=step_to_last_estimate,
            num_iterations=num_iterations,
            finished=finished,
        ),
        _BrentSearchParams(
            objective_fn=objective_fn,
            max_iterations=max_iterations,
            absolute_root_tolerance=absolute_root_tolerance,
            relative_root_tolerance=relative_root_tolerance,
            function_tolerance=function_tolerance,
            stopping_policy_fn=stopping_policy_fn,
        ),
        _BrentSearchConstants(false=false, zero=zero, zero_value=zero_value),
    )


# `_brent` currently only support inverse quadratic extrapolation.
# This will be fixed when adding the `brenth` variant.
def _brent(
    objective_fn,
    left_bracket,
    right_bracket,
    value_at_left_bracket=None,
    value_at_right_bracket=None,
    absolute_root_tolerance=2e-7,
    relative_root_tolerance=None,
    function_tolerance=2e-7,
    max_iterations=100,
    stopping_policy_fn=None,
    validate_args=False,
    name=None,
):
    r"""Finds root(s) of a function of a single variable using Brent's method.

    [Brent's method](https://en.wikipedia.org/wiki/Brent%27s_method) is a
    root-finding algorithm combining the bisection method, the secant method and
    extrapolation. Like bisection it is guaranteed to converge towards a root if
    one exists, but that convergence is superlinear and on par with less reliable
    methods.

    This implementation is a translation of the algorithm described in the
    [original article](https://academic.oup.com/comjnl/article/14/4/422/325237).

    Args:
      objective_fn: Python callable for which roots are searched. It must be a
        callable of a single `Tensor` parameter and return a `Tensor` of the same
        shape and dtype as `left_bracket`.
      left_bracket: `Tensor` or Python float representing the first starting
        points. The function will search for roots between each pair of points
        defined by `left_bracket` and `right_bracket`. The shape of `left_bracket`
        should match that of the input to `objective_fn`.
      right_bracket: `Tensor` of the same shape and dtype as `left_bracket` or
        Python float representing the second starting points. The function will
        search for roots between each pair of points defined by `left_bracket` and
        `right_bracket`. This argument must have the same shape as `left_bracket`.
      value_at_left_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `left_bracket`. If specified, this argument
        must have the same shape as `left_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      value_at_right_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `right_bracket`. If specified, this argument
        must have the same shape as `right_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      absolute_root_tolerance: Optional `Tensor` representing the absolute
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `2e-7`.
      relative_root_tolerance: Optional `Tensor` representing the relative
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `None` which translates to `4 *
          numpy.finfo(left_bracket.dtype.as_numpy_dtype).eps`.
      function_tolerance: Optional `Tensor` representing the tolerance used to
        check for roots. If the absolute value of `objective_fn` is smaller than
        or equal to `function_tolerance` at a given estimate, then that estimate
        is considered a root for the function. If specified, this argument must
        broadcast with the shape of `left_bracket` and have the same dtype. Set to
        zero to match Brent's original algorithm and to continue the search until
        an exact root is found.
        Default value: `2e-7`.
      max_iterations: Optional `Tensor` of an integral dtype or Python integer
        specifying the maximum number of steps to perform for each initial point.
        Must broadcast with the shape of `left_bracket`. If an element is set to
        zero, the function will not search for any root for the corresponding
        points in `left_bracket` and `right_bracket`. Instead, it will return the
        best estimate from the inputs.
        Default value: `100`.
      stopping_policy_fn: Python `callable` controlling the algorithm termination.
        It must be a callable accepting a `Tensor` of booleans with the shape of
        `left_bracket` (each denoting whether the search is finished for each
        starting point), and returning a scalar boolean `Tensor` (indicating
        whether the overall search should stop). Typical values are
        `tf.reduce_all` (which returns only when the search is finished for all
        pairs of points), and `tf.reduce_any` (which returns as soon as the search
        is finished for any pair of points).
        Default value: `None` which translates to `tf.reduce_all`.
      validate_args: Python `bool` indicating whether to validate arguments such
        as `left_bracket`, `right_bracket`, `absolute_root_tolerance`,
        `relative_root_tolerance`, `function_tolerance`, and `max_iterations`.
        Default value: `False`.
      name: Python `str` name prefixed to ops created by this function.

    Returns:
      brent_results: A Python object containing the following attributes:
        estimated_root: `Tensor` containing the best estimate explored. If the
          search was successful within the specified tolerance, this estimate is
          a root of the objective function.
        objective_at_estimated_root: `Tensor` containing the value of the
          objective function at `estimated_root`. If the search was successful
          within the specified tolerance, then this is close to 0. It has the
          same dtype and shape as `estimated_root`.
        num_iterations: `Tensor` containing the number of iterations performed.
          It has the same dtype as `max_iterations` and shape as `estimated_root`.
        converged: Scalar boolean `Tensor` indicating whether `estimated_root` is
          a root within the tolerance specified for the search. It has the same
          shape as `estimated_root`.

    Raises:
      ValueError: if the `stopping_policy_fn` is not callable.
    """

    with tf.name_scope(name or "brent_root"):
        state, params, constants = _prepare_brent_args(
            objective_fn,
            left_bracket,
            right_bracket,
            value_at_left_bracket,
            value_at_right_bracket,
            absolute_root_tolerance,
            relative_root_tolerance,
            function_tolerance,
            max_iterations,
            stopping_policy_fn,
        )

        assertions = []
        if validate_args:
            assertions += [
                tf.Assert(
                    tf.reduce_all(
                        state.value_at_last_estimate * state.value_at_best_estimate
                        <= constants.zero_value
                    ),
                    [state.value_at_last_estimate, state.value_at_best_estimate],
                ),
                tf.Assert(
                    tf.reduce_all(params.absolute_root_tolerance > constants.zero),
                    [params.absolute_root_tolerance],
                ),
                tf.Assert(
                    tf.reduce_all(params.relative_root_tolerance > constants.zero),
                    [params.relative_root_tolerance],
                ),
                tf.Assert(
                    tf.reduce_all(params.function_tolerance >= constants.zero),
                    [params.function_tolerance],
                ),
                tf.Assert(
                    tf.reduce_all(params.max_iterations >= state.num_iterations),
                    [params.max_iterations],
                ),
            ]

        with tf.compat.v1.control_dependencies(assertions):
            result = tf.while_loop(
                # Negate `_should_stop` to determine if the search should continue.
                # This means, in particular, that tf.reduce_*all* will return only
                # when the search is finished for *all* starting points.
                lambda loop_vars: ~_should_stop(loop_vars, params.stopping_policy_fn),
                lambda state: _brent_loop_body(state, params, constants),
                loop_vars=[state],
                maximum_iterations=max_iterations,
            )

    state = result[0]
    converged = tf.math.abs(state.value_at_best_estimate) <= function_tolerance

    return BrentResults(
        estimated_root=state.best_estimate,
        objective_at_estimated_root=state.value_at_best_estimate,
        num_iterations=state.num_iterations,
        converged=converged,
    )


def brentq(
    objective_fn: Callable[[types.RealTensor], types.RealTensor],
    left_bracket: types.RealTensor,
    right_bracket: types.RealTensor,
    value_at_left_bracket: types.RealTensor = None,
    value_at_right_bracket: types.RealTensor = None,
    absolute_root_tolerance: types.RealTensor = 2e-7,
    relative_root_tolerance: types.RealTensor = None,
    function_tolerance: types.RealTensor = 2e-7,
    max_iterations: types.IntTensor = 100,
    stopping_policy_fn: Callable[[types.BoolTensor], types.BoolTensor] = None,
    validate_args: bool = False,
    name: str = None,
) -> BrentResults:
    r"""Finds root(s) of a function of single variable using Brent's method.

    [Brent's method](https://en.wikipedia.org/wiki/Brent%27s_method) is a
    root-finding algorithm combining the bisection method, the secant method and
    extrapolation. Like bisection it is guaranteed to converge towards a root if
    one exists, but that convergence is superlinear and on par with less reliable
    methods.

    This implementation is a translation of the algorithm described in the
    [original article](https://academic.oup.com/comjnl/article/14/4/422/325237).

    #### Examples

    ```python
    import tensorflow as tf
    import tf_quant_finance as tff

    # Example 1: Roots of a single function for two pairs of starting points.

    f = lambda x: 63 * x**5 - 70 * x**3 + 15 * x + 2
    x1 = tf.constant([-10, 1], dtype=tf.float64)
    x2 = tf.constant([10, -1], dtype=tf.float64)

    tf.math.brentq(objective_fn=f, left_bracket=x1, right_bracket=x2)
    # ==> BrentResults(
    #    estimated_root=array([-0.14823253, -0.14823253]),
    #    objective_at_estimated_root=array([3.27515792e-15, 0.]),
    #    num_iterations=array([11, 6]),
    #    converged=array([True, True]))

    tff.math.root_search.brentq(objective_fn=f,
                                left_bracket=x1,
                                right_bracket=x2,
                                stopping_policy_fn=tf.reduce_any)
    # ==> BrentResults(
    #    estimated_root=array([-2.60718234, -0.14823253]),
    #    objective_at_estimated_root=array([-6.38579115e+03, 2.39763764e-11]),
    #    num_iterations=array([7, 6]),
    #    converged=array([False, True]))
    ```

    # Example 2: Roots of a multiplex function for one pair of starting points.

    def f(x):
      return tf.constant([0., 63.], dtype=tf.float64) * x**5 \
          + tf.constant([5., -70.], dtype=tf.float64) * x**3 \
          + tf.constant([-3., 15.], dtype=tf.float64) * x \
          + 2

    x1 = tf.constant([-5, -5], dtype=tf.float64)
    x2 = tf.constant([5, 5], dtype=tf.float64)

    tff.math.root_search.brentq(objective_fn=f, left_bracket=x1, right_bracket=x2)
    # ==> BrentResults(
    #    estimated_root=array([-1., -0.14823253]),
    #    objective_at_estimated_root=array([0., 2.08721929e-14]),
    #    num_iterations=array([13, 11]),
    #    converged=array([True, True]))

    # Example 3: Roots of a multiplex function for two pairs of starting points.

    def f(x):
      return tf.constant([0., 63.], dtype=tf.float64) * x**5 \
          + tf.constant([5., -70.], dtype=tf.float64) * x**3 \
          + tf.constant([-3., 15.], dtype=tf.float64) * x \
          + 2

    x1 = tf.constant([[-5, -5], [10, 10]], dtype=tf.float64)
    x2 = tf.constant([[5, 5], [-10, -10]], dtype=tf.float64)

    tff.math.root_search.brentq(objective_fn=f, left_bracket=x1, right_bracket=x2)
    # ==> BrentResults(
    #    estimated_root=array([
    #        [-1, -0.14823253],
    #        [-1, -0.14823253]]),
    #    objective_at_estimated_root=array([
    #        [0., 2.08721929e-14],
    #        [0., 2.08721929e-14]]),
    #    num_iterations=array([
    #        [13, 11],
    #        [15, 11]]),
    #    converged=array([
    #        [True, True],
    #        [True, True]]))
    ```

    Args:
      objective_fn: Python callable for which roots are searched. It must be a
        callable of a single `Tensor` parameter and return a `Tensor` of the same
        shape and dtype as `left_bracket`.
      left_bracket: `Tensor` or Python float representing the first starting
        points. The function will search for roots between each pair of points
        defined by `left_bracket` and `right_bracket`. The shape of `left_bracket`
        should match that of the input to `objective_fn`.
      right_bracket: `Tensor` of the same shape and dtype as `left_bracket` or
        Python float representing the second starting points. The function will
        search for roots between each pair of points defined by `left_bracket` and
        `right_bracket`. This argument must have the same shape as `left_bracket`.
      value_at_left_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `left_bracket`. If specified, this argument
        must have the same shape as `left_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      value_at_right_bracket: Optional `Tensor` or Python float representing the
        value of `objective_fn` at `right_bracket`. If specified, this argument
        must have the same shape as `right_bracket`. If not specified, the value
        will be evaluated during the search.
        Default value: None.
      absolute_root_tolerance: Optional `Tensor` representing the absolute
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `2e-7`.
      relative_root_tolerance: Optional `Tensor` representing the relative
        tolerance for estimated roots, with the total tolerance being calculated
        as `(absolute_root_tolerance + relative_root_tolerance * |root|) / 2`. If
        specified, this argument must be positive, broadcast with the shape of
        `left_bracket` and have the same dtype.
        Default value: `None` which translates to `4 *
          numpy.finfo(left_bracket.dtype.as_numpy_dtype).eps`.
      function_tolerance: Optional `Tensor` representing the tolerance used to
        check for roots. If the absolute value of `objective_fn` is smaller than
        or equal to `function_tolerance` at a given estimate, then that estimate
        is considered a root for the function. If specified, this argument must
        broadcast with the shape of `left_bracket` and have the same dtype. Set to
        zero to match Brent's original algorithm and to continue the search until
        an exact root is found.
        Default value: `2e-7`.
      max_iterations: Optional `Tensor` of an integral dtype or Python integer
        specifying the maximum number of steps to perform for each initial point.
        Must broadcast with the shape of `left_bracket`. If an element is set to
        zero, the function will not search for any root for the corresponding
        points in `left_bracket` and `right_bracket`. Instead, it will return the
        best estimate from the inputs.
        Default value: `100`.
      stopping_policy_fn: Python `callable` controlling the algorithm termination.
        It must be a callable accepting a `Tensor` of booleans with the shape of
        `left_bracket` (each denoting whether the search is finished for each
        starting point), and returning a scalar boolean `Tensor` (indicating
        whether the overall search should stop). Typical values are
        `tf.reduce_all` (which returns only when the search is finished for all
        pairs of points), and `tf.reduce_any` (which returns as soon as the search
        is finished for any pair of points).
        Default value: `None` which translates to `tf.reduce_all`.
      validate_args: Python `bool` indicating whether to validate arguments such
        as `left_bracket`, `right_bracket`, `absolute_root_tolerance`,
        `relative_root_tolerance`, `function_tolerance`, and `max_iterations`.
        Default value: `False`.
      name: Python `str` name prefixed to ops created by this function.

    Returns:
      brent_results: A Python object containing the following attributes:
        estimated_root: `Tensor` containing the best estimate explored. If the
          search was successful within the specified tolerance, this estimate is
          a root of the objective function.
        objective_at_estimated_root: `Tensor` containing the value of the
          objective function at `estimated_root`. If the search was successful
          within the specified tolerance, then this is close to 0. It has the
          same dtype and shape as `estimated_root`.
        num_iterations: `Tensor` containing the number of iterations performed.
          It has the same dtype as `max_iterations` and shape as `estimated_root`.
        converged: Scalar boolean `Tensor` indicating whether `estimated_root` is
          a root within the tolerance specified for the search. It has the same
          shape as `estimated_root`.

    Raises:
      ValueError: if the `stopping_policy_fn` is not callable.
    """

    return _brent(
        objective_fn,
        left_bracket,
        right_bracket,
        value_at_left_bracket=value_at_left_bracket,
        value_at_right_bracket=value_at_right_bracket,
        absolute_root_tolerance=absolute_root_tolerance,
        relative_root_tolerance=relative_root_tolerance,
        function_tolerance=function_tolerance,
        max_iterations=max_iterations,
        stopping_policy_fn=stopping_policy_fn,
        validate_args=validate_args,
        name=name,
    )



"""Basic PDFs are provided here.

Gauss, exponential... that can be used together with Functors to build larger models.
"""
#  Copyright (c) 2024 zfit

from __future__ import annotations

#  Copyright (c) 2023 zfit
import contextlib

from typing import Literal

import tensorflow as tf
from pydantic import Field

import zfit.z.numpy as znp
from zfit import z
from ..core.basepdf import BasePDF
from ..core.serialmixin import SerializableMixin
from ..core.space import ANY_LOWER, ANY_UPPER, Space
from ..serialization import SpaceRepr, Serializer
from ..serialization.pdfrepr import BasePDFRepr
from ..util import ztyping
from ..util.exception import BreakingAPIChangeError
from ..util.warnings import warn_advanced_feature
from ..util.ztyping import ExtendedInputType, NormInputType


class Exponential(BasePDF, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        lam=None,
        obs: ztyping.ObsTypeInput = None,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "Exponential",
        lambda_=None,
    ):
        """Exponential function exp(lambda * x).

        The function is normalized over a finite range and therefore a pdf. So the PDF is precisely
        defined as :math:`\\frac{ e^{\\lambda \\cdot x}}{ \\int_{lower}^{upper} e^{\\lambda \\cdot x} dx}`

        Args:
            lam: Lambda parameter of the exponential.
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|
        """
        if lambda_ is not None:
            if lam is None:
                lam = lambda_
            else:
                raise BreakingAPIChangeError(
                    "The 'lambda' parameter has been renamed from 'lambda_' to 'lam'."
                )
        params = {"lambda": lam, "lam": lam}
        super().__init__(obs, name=name, params=params, extended=extended, norm=norm)

        self._calc_numerics_data_shift = lambda: z.constant(0.0)

        if not self.space.has_limits:
            warn_advanced_feature(
                "Exponential pdf relies on a shift of the input towards 0 to keep the numerical "
                f"stability high. The space {self.space} does not have limits set and no shift"
                f" will occure. To set it manually, set _numerics_data_shift to the expected"
                f" average values given to this function _in case you want things to be set_."
                f"If this sounds unfamiliar, regard this as an error and use a normalization range.",
                identifier="exp_shift",
            )
        self._set_numerics_data_shift(self.space)

    def _unnormalized_pdf(self, x):
        lambda_ = self.params["lambda"].value()
        x = x.unstack_x()
        probs = znp.exp(lambda_ * (self._shift_x(x)))
        tf.debugging.assert_all_finite(
            probs,
            f"Exponential PDF {self} has non valid values. This is likely caused"
            f" by numerical problems: if the exponential is too steep, this will"
            f" yield NaNs or infs. Make sure that your lambda is small enough and/or"
            f" the initial space is in the same"
            f" region as your data (and norm, if explicitly set differently)."
            f" If this issue still persists, please oben an issue on Github:"
            f" https://github.com/zfit/zfit",
        )
        return probs  # Don't use exp! will overflow.

    def _shift_x(self, x):
        return x - self._calc_numerics_data_shift()

    @contextlib.contextmanager
    def _set_numerics_data_shift(self, limits):
        if limits:

            def calc_numerics_data_shift():
                lower, upper = [], []
                for limit in limits:
                    low, up = limit.rect_limits
                    lower.append(z.convert_to_tensor(low[:, 0]))
                    upper.append(z.convert_to_tensor(up[:, 0]))
                lower = z.convert_to_tensor(lower)
                upper = z.convert_to_tensor(upper)
                lower_val = znp.min(lower, axis=0)
                upper_val = znp.max(upper, axis=0)

                return (upper_val + lower_val) / 2

            old_value = self._calc_numerics_data_shift

            self._calc_numerics_data_shift = calc_numerics_data_shift
            yield
            self._calc_numerics_data_shift = old_value
        else:
            yield

    # All hooks are needed to set the right shift when "entering" the pdf. The norm range is taken where both are
    # available. No special need needs to be taken for sampling (it samples from the correct region, the limits, and
    # uses the predictions by the `unnormalized_prob` -> that is shifted correctly
    def _single_hook_integrate(self, limits, norm, x, options):
        with self._set_numerics_data_shift(norm):
            return super()._single_hook_integrate(limits, norm, x=x, options=None)

    def _single_hook_analytic_integrate(self, limits, norm):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_analytic_integrate(limits, norm)

    def _single_hook_numeric_integrate(self, limits, norm, options):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_numeric_integrate(limits, norm, options)

    def _single_hook_partial_integrate(self, x, limits, norm, *, options):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_partial_integrate(
                x, limits, norm, options=options
            )

    def _single_hook_partial_analytic_integrate(self, x, limits, norm):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_partial_analytic_integrate(x, limits, norm)

    def _single_hook_partial_numeric_integrate(self, x, limits, norm):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_partial_numeric_integrate(x, limits, norm)

    # def _single_hook_normalization(self, limits):
    #     with self._set_numerics_data_shift(limits=limits):
    #         return super()._single_hook_normalization(limits)

    #
    # # TODO: remove component_norm_range? But needed for integral?
    # def _single_hook_unnormalized_pdf(self, x, name):
    #     if component_norm_range.limits_are_false:
    #         component_norm_range = self.space
    #     if component_norm_range.limits_are_set:
    #         with self._set_numerics_data_shift(limits=component_norm_range):
    #             return super()._single_hook_unnormalized_pdf(x, name)
    #     else:
    #         return super()._single_hook_unnormalized_pdf(x, name)
    #
    def _single_hook_pdf(self, x, norm):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_pdf(x, norm)

    #
    def _single_hook_log_pdf(self, x, norm):
        with self._set_numerics_data_shift(limits=norm):
            return super()._single_hook_log_pdf(x, norm)

    def _single_hook_sample(self, n, limits, x=None):
        with self._set_numerics_data_shift(limits=limits):
            return super()._single_hook_sample(n, limits, x)


def _exp_integral_from_any_to_any(limits, params, model):
    lambda_ = params["lambda"]
    lower, upper = limits.rect_limits
    # if any(np.isinf([lower, upper])):
    #     raise AnalyticIntegralNotImplemented

    integral = _exp_integral_func_shifting(
        lambd=lambda_, lower=lower, upper=upper, model=model
    )
    return integral[0]


def _exp_integral_func_shifting(lambd, lower, upper, model):
    def raw_integral(x):
        return (
            z.exp(lambd * (model._shift_x(x))) / lambd
        )  # needed due to overflow in exp otherwise

    lower = z.convert_to_tensor(lower)
    lower_int = raw_integral(x=lower)
    upper = z.convert_to_tensor(upper)
    upper_int = raw_integral(x=upper)
    integral = upper_int - lower_int
    return integral


def exp_icdf(x, params, model):
    lambd = params["lambda"]
    x = z.unstack_x(x)
    x = model._shift_x(x)
    return znp.log(lambd * x) / lambd


# Exponential.register_inverse_analytic_integral(exp_icdf)  # TODO: register icdf for exponential
# TODO: cleanup, make cdf registrable _and_ inverse integral, but real

limits = Space(axes=0, limits=(ANY_LOWER, ANY_UPPER))
Exponential.register_analytic_integral(
    func=_exp_integral_from_any_to_any, limits=limits
)


class ExponentialPDFRepr(BasePDFRepr):
    _implementation = Exponential
    hs3_type: Literal["Exponential"] = Field("Exponential", alias="type")
    x: SpaceRepr
    lam: Serializer.types.ParamTypeDiscriminated



#  Copyright (c) 2023 zfit

from __future__ import annotations

import abc
import collections
from collections import OrderedDict
from collections.abc import Callable
from typing import Mapping, Iterable, List

import numpy as np
import pydantic

from .serialmixin import SerializableMixin
from ..serialization.serializer import BaseRepr, Serializer

from typing import Literal

import tensorflow as tf
import tensorflow_probability as tfp
from ordered_set import OrderedSet

import zfit.z.numpy as znp
from zfit import z
from .baseobject import BaseNumeric
from .dependents import _extract_dependencies
from .interfaces import ZfitConstraint, ZfitParameter
from ..settings import ztypes
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import ShapeIncompatibleError

tfd = tfp.distributions


# TODO(serialization): add to serializer
class BaseConstraintRepr(BaseRepr):
    _implementation = None
    _owndict = pydantic.PrivateAttr(default_factory=dict)
    hs3_type: Literal["BaseConstraint"] = pydantic.Field("BaseConstraint", alias="type")


class BaseConstraint(ZfitConstraint, BaseNumeric):
    def __init__(
        self,
        params: dict[str, ZfitParameter] = None,
        name: str = "BaseConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints.

        Args:
            dtype: the dtype of the constraint
            name: the name of the constraint
            params: A dictionary with the internal name of the
                parameter and the parameters itself the constrains depends on
        """
        super().__init__(name=name, dtype=dtype, params=params, **kwargs)

    def value(self):
        return self._value()

    @abc.abstractmethod
    def _value(self):
        raise NotImplementedError

    def _get_dependencies(self) -> ztyping.DependentsType:
        return _extract_dependencies(self.get_params(floating=None))


# TODO: improve arbitrary constraints, should we allow only functions that have a `params` argument?
class SimpleConstraint(BaseConstraint):
    def __init__(
        self,
        func: Callable,
        params: Mapping[str, ztyping.ParameterType]
        | Iterable[ztyping.ParameterType]
        | ztyping.ParameterType
        | None,
        *,
        name: str = None,
    ):
        """Constraint from a (function returning a) Tensor.

        Args:
            func: Callable that constructs the constraint and returns a tensor. For the expected signature,
                see below in ``params``.
            params: The parameters of the loss. If given as a list, the parameters are named "param_{i}"
                and the function does not take any arguments. If given as a dict, the function expects
                the parameter as the first argument (``params``).
        """
        if name is None:
            name = "SimpleConstraint"
        self._simple_func = func
        self._func_params = None
        if isinstance(params, collections.abc.Mapping):
            self._func_params = params
            params = list(params.values())
        self._simple_func_dependents = convert_to_container(
            params, container=OrderedSet
        )

        params = convert_to_container(params, container=list)
        if self._func_params is None:
            params = OrderedDict((f"param_{i}", p) for i, p in enumerate(params))
        else:
            params = self._func_params

        super().__init__(name=name, params=params)

    def _value(self):
        if self._func_params is None:
            return self._simple_func()
        else:
            return self._simple_func(self._func_params)


class ProbabilityConstraint(BaseConstraint):
    def __init__(
        self,
        observation: ztyping.NumericalScalarType | ZfitParameter,
        params: dict[str, ZfitParameter] = None,
        name: str = "ProbabilityConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints using a probability density function.

        Args:
            dtype: the dtype of the constraint
            name: the name of the constraint
            params: The parameters to constraint
            observation: Observed values of the parameter
                to constraint obtained from auxiliary measurements.
        """
        # TODO: proper handling of input params, arrays. ArrayParam?
        params = convert_to_container(params)
        params_dict = {f"param_{i}": p for i, p in enumerate(params)}
        super().__init__(name=name, dtype=dtype, params=params_dict, **kwargs)
        params = tuple(self.params.values())

        observation = convert_to_container(observation, tuple)
        if len(observation) != len(params):
            raise ShapeIncompatibleError(
                "observation and params have to be the same length. Currently"
                f"observation: {len(observation)}, params: {len(params)}"
            )

        self._observation = observation  # TODO: needed below? Why?
        # for obs, p in zip(observation, params):
        #     obs = convert_to_parameter(obs, f"{p.name}_obs", prefer_constant=False)
        #     obs.floating = False
        #     self._observation.append(obs)

        self._ordered_params = params

    @property
    def observation(self):
        """Return the observed values of the parameters constrained."""
        return self._observation

    def value(self):
        return self._value()

    @abc.abstractmethod
    def _value(self):
        raise NotImplementedError

    def _get_dependencies(self) -> ztyping.DependentsType:
        return _extract_dependencies(self.get_params())

    def sample(self, n):
        """Sample ``n`` points from the probability density function for the observed value of the parameters.

        Args:
            n: The number of samples to be generated.
        Returns:
        """
        sample = self._sample(n=n)
        return {p: sample[:, i] for i, p in enumerate(self._ordered_params)}

    @abc.abstractmethod
    def _sample(self, n):
        raise NotImplementedError

    @property
    def _params_array(self):
        return znp.asarray(self._ordered_params)


class TFProbabilityConstraint(ProbabilityConstraint):
    def __init__(
        self,
        observation: ztyping.NumericalScalarType | ZfitParameter,
        params: dict[str, ZfitParameter],
        distribution: tfd.Distribution,
        dist_params,
        dist_kwargs=None,
        name: str = "DistributionConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints using a probability density function from ``tensorflow_probability``.

        Args:
            distribution: The probability density function
                used to constraint the parameters
        """
        super().__init__(
            observation=observation, params=params, name=name, dtype=dtype, **kwargs
        )

        self._distribution = distribution
        self.dist_params = dist_params
        self.dist_kwargs = dist_kwargs if dist_kwargs is not None else {}

    @property
    def distribution(self):
        params = self.dist_params
        if callable(params):
            params = params(self.observation)
        kwargs = self.dist_kwargs
        if callable(kwargs):
            kwargs = kwargs()
        params = {k: tf.cast(v, ztypes.float) for k, v in params.items()}
        return self._distribution(**params, **kwargs, name=f"{self.name}_tfp")

    def _value(self):
        array = tf.cast(self._params_array, ztypes.float)
        value = -self.distribution.log_prob(array)
        return tf.reduce_sum(value)

    def _sample(self, n):
        return self.distribution.sample(n)


class GaussianConstraint(TFProbabilityConstraint, SerializableMixin):
    def __init__(
        self,
        params: ztyping.ParamTypeInput,
        observation: ztyping.NumericalScalarType,
        uncertainty: ztyping.NumericalScalarType,
    ):
        r"""Gaussian constraints on a list of parameters to some observed values with uncertainties.

        A Gaussian constraint is defined as the likelihood of ``params`` given the ``observations`` and ``uncertainty`` from
        a different measurement.

        .. math::
            \text{constraint} = \text{Gauss}(\text{observation}; \text{params}, \text{uncertainty})


        Args:
            params: The parameters to constraint; corresponds to x in the Gaussian
                distribution.
            observation: observed values of the parameter; corresponds to mu
                in the Gaussian distribution.
            uncertainty: Uncertainties or covariance/error
                matrix of the observed values. Can either be a single value, a list of values, an array or a tensor.
                Corresponds to the sigma of the Gaussian distribution.
        Raises:
            ShapeIncompatibleError: If params, mu and sigma have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple)
        params = convert_to_container(params, tuple)
        uncertainty = convert_to_container(uncertainty, tuple)
        if (
            isinstance(uncertainty[0], (np.ndarray, tf.Tensor))
            and len(uncertainty) == 1
        ):
            uncertainty = tuple(uncertainty[0])
        original_init = {
            "observation": observation,
            "params": params,
            "uncertainty": uncertainty,
        }

        def create_covariance(mu, sigma):
            mu = z.convert_to_tensor(mu)
            sigma = znp.asarray(
                sigma
            )  # otherwise TF complains that the shape got changed from [2] to [2, 2] (if we have a tuple of two arrays)
            sigma = z.convert_to_tensor(sigma)  # TODO (Mayou36): fix as above?
            params_tensor = z.convert_to_tensor(params)

            if sigma.shape.ndims > 1:
                covariance = sigma  # TODO: square as well?
            elif sigma.shape.ndims == 1:
                covariance = tf.linalg.tensor_diag(z.pow(sigma, 2.0))
            else:
                sigma = znp.reshape(sigma, [1])
                covariance = tf.linalg.tensor_diag(z.pow(sigma, 2.0))

            if (
                not params_tensor.shape[0]
                == mu.shape[0]
                == covariance.shape[0]
                == covariance.shape[1]
            ):
                raise ShapeIncompatibleError(
                    f"params_tensor, observation and uncertainty have to have the"
                    " same length. Currently"
                    f"param: {params_tensor.shape[0]}, mu: {mu.shape[0]}, "
                    f"covariance (from uncertainty): {covariance.shape[0:2]}"
                )
            return covariance

        distribution = tfd.MultivariateNormalTriL
        covariance = create_covariance(observation, uncertainty)
        dist_params = lambda observation: dict(
            loc=observation,
            scale_tril=tf.linalg.cholesky(covariance),
        )
        dist_kwargs = dict(validate_args=True)

        super().__init__(
            name="GaussianConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)
        self._covariance = lambda: create_covariance(self.observation, uncertainty)

    @property
    def covariance(self):
        """Return the covariance matrix of the observed values of the parameters constrained."""
        return self._covariance()


class GaussianConstraintRepr(BaseConstraintRepr):
    _implementation = GaussianConstraint
    hs3_type: Literal["GaussianConstraint"] = pydantic.Field(
        "GaussianConstraint", alias="type"
    )

    params: List[Serializer.types.ParamInputTypeDiscriminated]
    observation: List[Serializer.types.ParamInputTypeDiscriminated]
    uncertainty: List[Serializer.types.ParamInputTypeDiscriminated]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation", "uncertainty")
    def validate_params(cls, v):
        if isinstance(v, np.ndarray):
            v = v.tolist()
        else:
            v = convert_to_container(v, list)
        return v


class PoissonConstraint(TFProbabilityConstraint, SerializableMixin):
    def __init__(
        self, params: ztyping.ParamTypeInput, observation: ztyping.NumericalScalarType
    ):
        r"""Poisson constraints on a list of parameters to some observed values.

        Constraints parameters that can be counts (i.e. from a histogram) or, more generally, are
        Poisson distributed. This is often used in the case of histogram templates which are obtained
        from simulation and have a poisson uncertainty due to limited statistics.

        .. math::
            \text{constraint} = \text{Poisson}(\text{observation}; \text{params})


        Args:
            params: The parameters to constraint; corresponds to the mu in the Poisson
                distribution.
            observation: observed values of the parameter; corresponds to lambda
                in the Poisson distribution.
        Raises:
            ShapeIncompatibleError: If params and observation have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple)
        params = convert_to_container(params, tuple)
        original_init = {"observation": observation, "params": params}

        distribution = tfd.Poisson
        dist_params = dict(rate=observation)
        dist_kwargs = dict(validate_args=False)

        super().__init__(
            name="PoissonConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)


class PoissonConstraintRepr(BaseConstraintRepr):
    _implementation = PoissonConstraint
    hs3_type: Literal["PoissonConstraint"] = pydantic.Field(
        "PoissonConstraint", alias="type"
    )

    params: List[Serializer.types.ParamInputTypeDiscriminated]
    observation: List[Serializer.types.ParamInputTypeDiscriminated]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation")
    def validate_params(cls, v):
        if isinstance(v, np.ndarray):
            v = v.tolist()
        else:
            v = convert_to_container(v, list)
        return v


class LogNormalConstraint(TFProbabilityConstraint, SerializableMixin):
    def __init__(
        self,
        params: ztyping.ParamTypeInput,
        observation: ztyping.NumericalScalarType,
        uncertainty: ztyping.NumericalScalarType,
    ):
        r"""Log-normal constraints on a list of parameters to some observed values.

        Constraints parameters that can be counts (i.e. from a histogram) or, more generally, are
        LogNormal distributed. This is often used in the case of histogram templates which are obtained
        from simulation and have a log-normal uncertainty due to a multiplicative uncertainty.

        .. math::
            \text{constraint} = \text{LogNormal}(\text{observation}; \text{params})


        Args:
            params: The parameters to constraint; corresponds to the mu in the Poisson
                distribution.
            observation: observed values of the parameter; corresponds to lambda
                in the Poisson distribution.
            uncertainty: uncertainty of the observed values of the parameter; corresponds to sigma
                in the Poisson distribution.
        Raises:
            ShapeIncompatibleError: If params, mu and sigma have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple)
        params = convert_to_container(params, tuple)
        uncertainty = convert_to_container(uncertainty, tuple)
        original_init = {
            "observation": observation,
            "params": params,
            "uncertainty": uncertainty,
        }

        distribution = tfd.LogNormal
        dist_params = lambda observation: dict(loc=observation, scale=uncertainty)
        dist_kwargs = dict(validate_args=False)

        super().__init__(
            name="LogNormalConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)


class LogNormalConstraintRepr(BaseConstraintRepr):
    _implementation = LogNormalConstraint
    hs3_type: Literal["LogNormalConstraint"] = pydantic.Field(
        "LogNormalConstraint", alias="type"
    )

    params: List[Serializer.types.ParamInputTypeDiscriminated]
    observation: List[Serializer.types.ParamInputTypeDiscriminated]
    uncertainty: List[Serializer.types.ParamInputTypeDiscriminated]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation", "uncertainty")
    def validate_params(cls, v):
        if isinstance(v, np.ndarray):
            v = v.tolist()
        else:
            v = convert_to_container(v, list)
        return v



#  Copyright (c) 2023 zfit

from __future__ import annotations


def make_param_constructor(constructor):
    """Create a constructor for a parameter class avoiding the `NameAlreadyTakenError`.

    If the parameter already exists, it is returned instead of creating a new one.

    Args:
        constructor: Callable that creates the parameter.

    Returns:
        Callable that creates the parameter.
    """

    def param_constructor(name, **kwargs):
        from ..core.parameter import Parameter

        if name in Parameter._existing_params:
            return Parameter._existing_params[name]
        else:
            return constructor(name=name, **kwargs)

    return param_constructor



#  Copyright (c) 2023 zfit
import tensorflow_probability as tfp
from zfit_interface.data import ZfitData


@tfp.experimental.auto_composite_tensor()
class UnbinnedData(tfp.experimental.AutoCompositeTensor, ZfitData):
    def __init__(self, data, space=None, weights=None):
        self._data = data
        self._space = space
        self._weights = weights

    @property
    def is_binned(self):
        return False

    @property
    def is_unbinned(self):
        return True

    @property
    def space(self):
        return self._space

    @property
    def data(self):
        return self._data

    @property
    def weights(self):
        return self._weights

    def values(self):
        return self.data

    def __getitem__(self, item):
        if not isinstance(item, str):
            return super().__getitem__(item)
        for index, axis in enumerate(self.axes):
            if axis.name == item:
                break
        else:
            raise KeyError(f"{item} not in {self.axes}")
        return self.data[..., index]

    @property
    def has_weights(self):
        return self._weights is not None


# tensorlike.register_tensor_conversion(UnbinnedData, name='UnbinnedData', overload_operators=True)



#  Copyright (c) 2022 zfit
import hist
import numpy as np
import pytest

import zfit
import zfit.z.numpy as znp
from zfit.data import BinnedData
from zfit.util.exception import NotExtendedPDFError


def test_sum_histogram_pdf():
    bins1 = 5
    bins2 = 7
    counts = znp.random.uniform(high=1, size=(bins1, bins2))  # generate counts
    counts2 = np.random.normal(loc=5, size=(bins1, bins2))
    counts3 = (
        znp.linspace(0, 10, num=bins1)[:, None] * znp.linspace(0, 5, num=bins2)[None, :]
    )
    binnings = [
        zfit.binned.RegularBinning(bins1, 0, 10, name="obs1"),
        zfit.binned.RegularBinning(7, -10, bins2, name="obs2"),
    ]
    binning = binnings
    obs = zfit.Space(obs=["obs1", "obs2"], binning=binning)

    data = BinnedData.from_tensor(
        space=obs, values=counts, variances=znp.ones_like(counts) * 1.3
    )
    data2 = BinnedData.from_tensor(obs, counts2)
    data3 = BinnedData.from_tensor(obs, counts3)

    pdf = zfit.pdf.HistogramPDF(data=data, extended=znp.sum(counts))
    pdf2 = zfit.pdf.HistogramPDF(data=data2, extended=znp.sum(counts2))
    pdf3 = zfit.pdf.HistogramPDF(data=data3, extended=znp.sum(counts3))
    assert len(pdf.ext_pdf(data)) > 0
    pdf_sum = zfit.pdf.BinnedSumPDF(pdfs=[pdf, pdf2, pdf3], obs=obs)

    probs = pdf_sum.counts(data)
    true_sum_counts = counts + counts2 + counts3
    np.testing.assert_allclose(true_sum_counts, probs)
    nsamples = 100_000_000
    sample = pdf_sum.sample(n=nsamples)
    np.testing.assert_allclose(
        true_sum_counts, sample.values() / nsamples * pdf_sum.get_yield(), rtol=0.03
    )

    # integrate
    true_integral = znp.sum(true_sum_counts)
    integral = pdf_sum.ext_integrate(limits=obs)
    assert pytest.approx(float(true_integral)) == float(integral)


def test_pdf_formhist():
    h = hist.Hist(
        hist.axis.Regular(3, -3, 3, name="x", flow=False),
        hist.axis.Regular(2, -5, 5, name="y", flow=False),
        storage=hist.storage.Weight(),
    )

    x2 = np.random.randn(1_000)
    y2 = 0.5 * np.random.randn(1_000)

    h.fill(x=x2, y=y2)
    pdf = zfit.pdf.HistogramPDF(data=h, extended=False)
    assert not pdf.is_extended
    ntot = h.sum().value
    np.testing.assert_allclose(h.values() / ntot, pdf.rel_counts(h))
    with pytest.raises(NotExtendedPDFError):
        _ = pdf.counts()

    np.testing.assert_allclose(h.density(), pdf.pdf(h))
    with pytest.raises(NotExtendedPDFError):
        _ = pdf.ext_pdf(h)

    # test extended
    ext_pdf = zfit.pdf.HistogramPDF(data=h, extended=True)
    assert ext_pdf.is_extended

    # test pdf
    np.testing.assert_allclose(h.values() / ntot, ext_pdf.rel_counts(h))
    np.testing.assert_allclose(h.counts(), ext_pdf.counts(h))
    # test counts
    np.testing.assert_allclose(h.density(), pdf.pdf(h))
    np.testing.assert_allclose(h.density() * ntot, ext_pdf.ext_pdf(h))

    # test sample
    sample = ext_pdf.sample(n=1_000)
    assert sample.nevents == 1_000
    assert sample.rank == 2

    sample = pdf.sample(n=1_000)
    assert sample.nevents == 1_000
    assert sample.rank == 2

    # test integral
    limits = ext_pdf.space
    assert pytest.approx(ntot) == ext_pdf.ext_integrate(limits)
    assert pytest.approx(1.0) == float(ext_pdf.integrate(limits))
    assert pytest.approx(1.0) == float(pdf.integrate(limits))



#  Copyright (c) 2023 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

from zfit import z

if TYPE_CHECKING:
    import zfit

import numpy as np
import pytest
import tensorflow as tf

import zfit
from zfit.core.data import Data
from zfit.core.interfaces import ZfitPDF
from zfit.core.parameter import Parameter
from zfit.core.space import Space
from zfit.models.dist_tfp import Gauss
from zfit.models.functor import ProductPDF, SumPDF

low, high = -0.64, 5.9

mu1_true = 1.0
mu2_true = 2.0
mu3_true = 0.6
sigma1_true = 1.4
sigma2_true = 2.3
sigma3_true = 1.8

fracs = [0.3, 0.15]

obs1 = zfit.Space("obs1", (-3, 6))
obs2 = zfit.Space("obs2", (-2, 5))
obs3 = zfit.Space("obs3", (-3.5, 4))
obs4 = zfit.Space("obs4", (-2.6, 4.1))


def true_gaussian_sum(x):
    def norm(sigma):
        return np.sqrt(2 * np.pi) * sigma

    sum_gauss = (
        fracs[0]
        * np.exp(-((x - mu1_true) ** 2) / (2 * sigma1_true**2))
        / norm(sigma1_true)
    )
    sum_gauss += (
        fracs[1]
        * np.exp(-((x - mu2_true) ** 2) / (2 * sigma2_true**2))
        / norm(sigma2_true)
    )
    sum_gauss += (
        (1.0 - sum(fracs))
        * np.exp(-((x - mu3_true) ** 2) / (2 * sigma3_true**2))
        / norm(sigma3_true)
    )
    return sum_gauss


def create_params(name_add=""):
    mu1 = Parameter("mu1" + name_add, mu1_true)
    mu2 = Parameter("mu2" + name_add, mu2_true)
    mu3 = Parameter("mu3" + name_add, mu3_true)
    sigma1 = Parameter("sigma1" + name_add, sigma1_true)
    sigma2 = Parameter("sigma2" + name_add, sigma2_true)
    sigma3 = Parameter("sigma3" + name_add, sigma3_true)
    return mu1, mu2, mu3, sigma1, sigma2, sigma3


def create_gaussians() -> list[ZfitPDF]:
    # Gauss for sum, same axes
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params()
    gauss1 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss1asum")
    gauss2 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss2asum")
    gauss3 = Gauss(mu=mu3, sigma=sigma3, obs=obs1, name="gauss3asum")
    return [gauss1, gauss2, gauss3]


def sum_gaussians():
    gauss_dists = create_gaussians()
    sum_gauss = SumPDF(pdfs=gauss_dists, fracs=fracs, obs=obs1)
    return sum_gauss


def product_gaussians():
    gauss_dists = create_gaussians()
    prod_gauss = ProductPDF(pdfs=gauss_dists, obs=obs1)
    return prod_gauss


def product_gauss_3d(name=""):
    # Gauss for product, independent
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params("3d" + name)

    gauss13 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss1a")
    gauss23 = Gauss(mu=mu2, sigma=sigma2, obs=obs2, name="gauss2a")
    gauss33 = Gauss(mu=mu3, sigma=sigma3, obs=obs3, name="gauss3a")
    gauss_dists3 = [gauss13, gauss23, gauss33]
    prod_gauss_3d = ProductPDF(pdfs=gauss_dists3)
    return prod_gauss_3d

    # prod_gauss_3d.update_integration_options(draws_per_dim=33)


def test_product_separation():
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params()
    gauss1 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss1asum")
    gauss2 = Gauss(mu=mu2, sigma=sigma2, obs=obs2, name="gauss2asum")
    gauss3 = Gauss(mu=mu3, sigma=sigma3, obs=obs1, name="gauss3asum")
    gauss4 = Gauss(mu=mu3, sigma=sigma3, obs=obs3, name="gauss3asum")
    gauss5 = Gauss(mu=mu3, sigma=sigma3, obs=obs4, name="gauss3asum")

    prod12 = ProductPDF(pdfs=[gauss1, gauss2, gauss3])
    assert not prod12._prod_is_same_obs_pdf
    prod13 = ProductPDF(pdfs=[gauss3, gauss4])
    assert not prod13._prod_is_same_obs_pdf
    prod123 = ProductPDF([prod12, prod13])
    assert prod123._prod_is_same_obs_pdf
    npoints = 2000
    data3 = zfit.Data.from_numpy(array=np.linspace(0, 1, npoints), obs=obs3)
    data2 = zfit.Data.from_numpy(array=np.linspace(0, 1, npoints), obs=obs2)
    data1 = zfit.Data.from_numpy(array=np.linspace(0, 1, npoints), obs=obs1)
    data4 = zfit.Data.from_numpy(array=np.linspace(0, 1, npoints), obs=obs4)
    integral13 = prod13.partial_integrate(x=data3, limits=obs1, norm=False)
    assert integral13.shape[0] == npoints
    trueint3 = gauss4.pdf(data3, norm=False) * gauss3.integrate(obs1, norm=False)
    np.testing.assert_allclose(integral13, trueint3)
    assert (
        prod12.partial_integrate(
            x=data2,
            limits=obs1,
        ).shape[0]
        == npoints
    )
    assert (
        prod123.partial_integrate(
            x=data3,
            limits=obs1 * obs2,
        ).shape[0]
        == npoints
    )
    assert (
        prod123.partial_integrate(
            x=data2,
            limits=obs1 * obs3,
        ).shape[0]
        == npoints
    )

    prod1234 = ProductPDF(pdfs=[gauss1, gauss2, gauss4, gauss5])
    integ = prod1234.partial_integrate(data1, limits=obs2 * obs3 * obs4, norm=False)
    assert integ.shape[0] == npoints

    obs13 = obs1 * obs3
    analytic_int = zfit.run(prod13.analytic_integrate(limits=obs13, norm=False))
    numeric_int = zfit.run(prod13.numeric_integrate(limits=obs13, norm=False))
    assert pytest.approx(analytic_int, rel=1e-3) == numeric_int


def product_gauss_4d():
    mu1, mu2, mu3, sigma1, sigma2, sigma3 = create_params("4d")

    obs4 = zfit.Space("obs4", (-4.5, 4.7))
    gauss12 = Gauss(mu=mu1, sigma=sigma1, obs=obs4, name="gauss12a")
    gauss22 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss22a")
    gauss32 = Gauss(mu=mu3, sigma=sigma3, obs=obs3, name="gauss32a")
    gauss_dists2 = [gauss12, gauss22, gauss32]
    obs = zfit.Space(["obs1", "obs2", "obs3", "obs4"])

    prod_3d = product_gauss_3d("4d")
    prod_gauss_4d = ProductPDF(pdfs=gauss_dists2 + [prod_3d], obs=obs)
    return prod_gauss_4d


def test_prod_gauss_nd():
    test_values = np.random.random(size=(10, 3))
    test_values_data = Data.from_tensor(obs=obs1 * obs2 * obs3, tensor=test_values)
    product_pdf = product_gauss_3d()
    assert len(product_pdf._prod_disjoint_obs_pdfs) == 3
    assert product_pdf._prod_is_same_obs_pdf is False
    assert product_pdf.n_obs == 3
    probs = product_pdf.pdf(x=test_values_data)
    gaussians = create_gaussians()
    for gauss, space in zip(gaussians, [obs1, obs2, obs3]):
        gauss.set_norm_range(space.rect_limits)
    true_probs = np.prod(
        [gauss.pdf(test_values[:, i]) for i, gauss in enumerate(gaussians)], axis=0
    )
    probs_np = probs.numpy()
    np.testing.assert_allclose(true_probs, probs_np, rtol=1e-2)


@pytest.mark.flaky(reruns=3)
def test_prod_gauss_nd_mixed():
    norm = (-5, 4)
    low, high = norm
    test_values = np.random.uniform(low=low, high=high, size=(1000, 4))

    obs4d = ["obs1", "obs2", "obs3", "obs4"]
    test_values_data = Data.from_tensor(obs=obs4d, tensor=test_values)
    limits_4d = Space(limits=(((-5,) * 4,), ((4,) * 4,)), obs=obs4d)
    prod_gauss_4d = product_gauss_4d()
    prod_gauss_4d.set_norm_range(limits_4d)
    probs = prod_gauss_4d.pdf(x=test_values_data, norm=limits_4d)
    gausses = create_gaussians()

    for gauss in gausses:
        gauss.set_norm_range(norm)
    gauss1, gauss2, gauss3 = gausses
    prod_gauss_3d = product_gauss_3d()

    def probs_4d(values):
        true_prob = [gauss1.pdf(values[:, 3])]
        true_prob += [gauss2.pdf(values[:, 0])]
        true_prob += [gauss3.pdf(values[:, 2])]
        true_prob += [
            prod_gauss_3d.pdf(
                values[:, 0:3],
                norm=Space(
                    limits=(((-5,) * 3,), ((4,) * 3,)), obs=["obs1", "obs2", "obs3"]
                ),
            )
        ]
        return tf.math.reduce_prod(true_prob, axis=0)

    true_unnormalized_probs = probs_4d(values=test_values)

    normalization_probs = limits_4d.area() * probs_4d(
        z.random.uniform(minval=low, maxval=high, shape=(40**4, 4))
    )
    true_probs = true_unnormalized_probs / tf.reduce_mean(
        input_tensor=normalization_probs
    )
    probs_np = probs.numpy()
    true_probs_np = true_probs.numpy()
    assert np.average(probs_np * limits_4d.area()) == pytest.approx(
        1.0, rel=0.33
    )  # low n mc
    np.testing.assert_allclose(true_probs_np, probs_np, rtol=2e-2)


def test_func_sum():
    sum_gauss = sum_gaussians()
    test_values = np.random.uniform(low=-3, high=4, size=10)
    sum_gauss_as_func = sum_gauss.as_func(norm=(-10, 10))
    vals = sum_gauss_as_func.func(x=test_values)
    vals = zfit.run(vals)
    # test_sum = sum([g.func(test_values) for g in gauss_dists])
    np.testing.assert_allclose(
        vals, true_gaussian_sum(test_values), rtol=3e-2
    )  # MC integral


def test_normalization_sum_gauss():
    normalization_testing(sum_gaussians())


def test_normalization_sum_gauss_extended():
    test_yield = 109.0
    sum_gauss_extended = sum_gaussians().create_extended(test_yield)
    normalization_testing(sum_gauss_extended)


def test_normalization_prod_gauss():
    normalization_testing(product_gaussians())


def test_exp():
    lambda_true = 0.031
    lambda_ = zfit.Parameter("lambda1", lambda_true)
    exp1 = zfit.pdf.Exponential(lam=lambda_, obs=zfit.Space("obs1", (-11, 11)))
    sample = exp1.sample(n=1000, limits=(-10, 10))
    sample_np = sample.numpy()
    assert not any(np.isnan(sample_np))

    exp2 = zfit.pdf.Exponential(lam=lambda_, obs=zfit.Space("obs1", (5250, 5750)))
    probs2 = exp2.pdf(x=np.linspace(5300, 5700, num=1100))
    probs2_np = probs2.numpy()
    assert not any(np.isnan(probs2_np))
    normalization_testing(exp2, limits=(5400, 5800))

    intlim = [5400, 5500]
    integral2 = zfit.run(
        exp2.integrate(
            intlim,
        )
    )
    numintegral2 = zfit.run(exp2.numeric_integrate(intlim))
    assert integral2 == pytest.approx(numintegral2, rel=0.03)


def normalization_testing(pdf, limits=None):
    limits = (low, high) if limits is None else limits
    space = Space(obs=obs1, limits=limits)
    with pdf.set_norm_range(space):
        samples = tf.cast(
            np.random.uniform(
                low=space.lower, high=space.upper, size=(40000, pdf.n_obs)
            ),
            dtype=tf.float64,
        )
        samples = zfit.Data.from_tensor(obs=space, tensor=samples)
        probs = pdf.pdf(samples)
        result = probs.numpy()
        result = zfit.run(np.average(result) * space.rect_area())
        assert pytest.approx(result, rel=0.03) == 1


def test_extended_gauss():
    mu1 = Parameter("mu11", 1.0)
    mu2 = Parameter("mu21", 12.0)
    mu3 = Parameter("mu31", 3.0)
    sigma1 = Parameter("sigma11", 1.0)
    sigma2 = Parameter("sigma21", 12.0)
    sigma3 = Parameter("sigma31", 33.0)
    yield1 = Parameter("yield11", 150.0)
    yield2 = Parameter("yield21", 550.0)
    yield3 = Parameter("yield31", 2500.0)

    gauss1 = Gauss(mu=mu1, sigma=sigma1, obs=obs1, name="gauss11")
    gauss2 = Gauss(mu=mu2, sigma=sigma2, obs=obs1, name="gauss21")
    gauss3 = Gauss(mu=mu3, sigma=sigma3, obs=obs1, name="gauss31")
    gauss1 = gauss1.create_extended(yield1)
    gauss2 = gauss2.create_extended(yield2)
    gauss3 = gauss3.create_extended(yield3)

    gauss_dists = [gauss1, gauss2, gauss3]

    sum_gauss = SumPDF(pdfs=gauss_dists)
    integral_true = (
        sum_gauss.integrate(
            (-1, 5),
        )
        * sum_gauss.get_yield()
    )

    assert zfit.run(integral_true) == pytest.approx(
        zfit.run(
            sum_gauss.ext_integrate(
                (-1, 5),
            )
        )
    )
    normalization_testing(pdf=sum_gauss, limits=obs1)


if __name__ == "__main__":
    test_prod_gauss_nd()



#  Copyright (c) 2022 zfit


from .minimizers.baseminimizer import (
    BaseMinimizer,
    BaseMinimizerV1,
    DefaultStrategy,
    minimize_supports,
)
from .minimizers.ipopt import IpyoptV1
from .minimizers.minimizer_minuit import Minuit
from .minimizers.minimizer_nlopt import (
    NLoptBaseMinimizerV1,
    NLoptBOBYQAV1,
    NLoptCCSAQV1,
    NLoptCOBYLAV1,
    NLoptESCHV1,
    NLoptISRESV1,
    NLoptLBFGSV1,
    NLoptMLSLV1,
    NLoptMMAV1,
    NLoptShiftVarV1,
    NLoptSLSQPV1,
    NLoptStoGOV1,
    NLoptSubplexV1,
    NLoptTruncNewtonV1,
)
from .minimizers.minimizers_scipy import (
    ScipyBaseMinimizerV1,
    ScipyLBFGSBV1,
    ScipyNelderMeadV1,
    ScipyNewtonCGV1,
    ScipyPowellV1,
    ScipySLSQPV1,
    ScipyTruncNCV1,
    ScipyTrustConstrV1,
)
from .minimizers.optimizers_tf import Adam, WrapOptimizer
from .minimizers.strategy import DefaultToyStrategy, PushbackStrategy, ZfitStrategy
from .minimizers.termination import EDM


class AdamMinimizer(Adam):
    def __init__(
        self,
        tol=None,
        learning_rate=0.2,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-08,
        name="Adam",
        **kwargs
    ):
        raise RuntimeError("Use zfit.minimize.Adam instead.")


class MinuitMinimizer(Minuit):
    def __init__(
        self,
        strategy: ZfitStrategy = None,
        minimize_strategy: int = 1,
        tol: float = None,
        verbosity: int = 5,
        name: str = None,
        ncall: int = 10000,
        use_minuit_grad: bool = None,
        **options
    ):
        raise RuntimeError(None, "Use zfit.minimize.Minuit instead.")


class Scipy:
    def __init__(self, *_, **__):
        raise RuntimeError(
            "This has been removed. Use the new Scipy* minimizer instead. In case anyone is missing,"
            " please feel free to open a request:"
            " https://github.com/zfit/zfit/issues/new?assignees="
            "&labels=discussion&template=feature-request.md&title="
            " or directly make a PR."
        )


class ScipyMinimizer(Scipy):
    pass


class BFGS:
    def __init__(self) -> None:
        raise RuntimeError(
            "BFGS (from TensorFlow Probability) has been removed as it is currently"
            " not working well. Use other BFGS-like implementations such as ScipyLBFGSBV1"
            " or NLoptLBFGSV1."
        )


__all__ = [
    "WrapOptimizer",
    "Adam",
    "Minuit",
    "ScipyBaseMinimizerV1",
    "ScipyLBFGSBV1",
    "ScipyTrustConstrV1",
    "ScipyPowellV1",
    "ScipySLSQPV1",
    "ScipyNewtonCGV1",
    "ScipyTruncNCV1",
    "ScipyNelderMeadV1",
    "NLoptBaseMinimizerV1",
    "NLoptLBFGSV1",
    "NLoptTruncNewtonV1",
    "NLoptSLSQPV1",
    "NLoptMMAV1",
    "NLoptCCSAQV1",
    "NLoptShiftVarV1",
    "NLoptMLSLV1",
    "NLoptStoGOV1",
    "NLoptESCHV1",
    "NLoptISRESV1",
    "NLoptSubplexV1",
    "NLoptBOBYQAV1",
    "NLoptCOBYLAV1",
    "IpyoptV1",
    "BaseMinimizer",
    "BaseMinimizerV1",
    "minimize_supports",
    "DefaultStrategy",
    "DefaultToyStrategy",
    "PushbackStrategy",
    "EDM",
]



#  Copyright (c) 2023 zfit
import json

# In[2]:
import zfit
import zfit.z.numpy as znp
from zfit.models.binned_functor import BinnedSumPDF
from zfit.models.template import BinnedTemplatePDFV1


def test_simple_examples_1D():
    import zfit.data
    import zfit.z.numpy as znp

    bkgnp = [50.0, 60.0]
    signp = [5.0, 10.0]
    datanp = [60.0, 80.0]
    uncnp = [5.0, 12.0]

    serialized = (
        """{
                            "channels": [
                                { "name": "singlechannel",
                                  "samples": [
                                    { "name": "signal",
                                    """
        + f"""

              "data": {signp},
              """
        """
                                                              "modifiers": [ { "name": "mu", "type": "normfactor", "data": null} ]
                                                            },
                                                            { "name": "background",
                                                            """
        f'"data": {bkgnp},'
        """
                                                              "modifiers": [ {"name": "uncorr_bkguncrt", "type": "shapesys",
                                                              """
        f'"data": {uncnp}'
        """
                                                          } ]
                                                        }
                                                      ]
                                                    }
                                                ],
                                                "observations": [
                                                    {
                                                    """
        f'"name": "singlechannel", "data": {datanp}'
        """
                                                            }
                                                        ],
                                                        "measurements": [
                                                            { "name": "Measurement", "config": {"poi": "mu", "parameters": []} }
                                                        ],
                                                        "version": "1.0.0"
                                                        }"""
    )

    obs = zfit.Space(
        "signal", binning=zfit.binned.RegularBinning(2, 0, 2, name="signal")
    )
    zdata = zfit.data.BinnedData.from_tensor(obs, datanp)
    zmcsig = zfit.data.BinnedData.from_tensor(obs, signp)
    zmcbkg = zfit.data.BinnedData.from_tensor(obs, bkgnp)

    shapesys = {
        f"shapesys_{i}": zfit.Parameter(f"shapesys_{i}", 1, 0.1, 10) for i in range(2)
    }
    bkgmodel = BinnedTemplatePDFV1(zmcbkg, sysshape=shapesys)
    # sigyield = zfit.Parameter('sigyield', znp.sum(zmcsig.values()))
    mu = zfit.Parameter("mu", 1, 0.1, 10)
    # sigmodeltmp = BinnedTemplatePDFV1(zmcsig)
    sigyield = zfit.ComposedParameter(
        "sigyield",
        lambda params: params["mu"] * znp.sum(zmcsig.values()),
        params={"mu": mu},
    )
    sigmodel = BinnedTemplatePDFV1(zmcsig, extended=sigyield)
    zmodel = BinnedSumPDF([sigmodel, bkgmodel])
    unc = np.array(uncnp) / np.array(bkgnp)
    nll = zfit.loss.ExtendedBinnedNLL(
        zmodel,
        zdata,
        constraints=zfit.constraint.GaussianConstraint(
            list(shapesys.values()), [1, 1], unc
        ),
    )
    # print(nll.value())
    # print(nll.gradient())
    # minimizer = zfit.minimize.ScipyLBFGSBV1()
    # minimizer = zfit.minimize.IpyoptV1()
    minimizer = zfit.minimize.Minuit(tol=1e-5, gradient=False)
    result = minimizer.minimize(nll)
    result.hesse(method="hesse_np")
    # result.errors()
    _ = str(result)
    # mu_z = sigmodel.get_yield() / znp.sum(zmcsig.values())
    zbestfit = np.asarray(result.params)
    errors = np.array([p["hesse"]["error"] for p in result.params.values()])
    # print('minval actual:', nll.value(), nll.gradient())
    # errors = np.ones(3) * 0.1
    # print('mu:', mu_z)

    spec = json.loads(serialized)

    workspace = pyhf.Workspace(spec)
    model = workspace.model(poi_name="mu")

    pars = model.config.suggested_init()
    data = workspace.data(model)

    model.logpdf(pars, data)

    bestfit_pars, twice_nll = pyhf.infer.mle.fit(data, model, return_fitted_val=True)
    diff = (bestfit_pars - zbestfit) / errors
    # print(bestfit_pars)
    np.testing.assert_allclose(diff, 0, atol=1e-3)

    # print(-2 * model.logpdf(bestfit_pars, data), twice_nll)


import pyhf
from pyhf.simplemodels import uncorrelated_background
import numpy as np
import pytest


def generate_source_static(n_bins):
    """Create the source structure for the given number of bins.

    Args:
        n_bins: `list` of number of bins

    Returns:
        source
    """
    binning = [n_bins, -0.5, n_bins + 0.5]
    data = [120.0] * n_bins
    bkg = [100.0] * n_bins
    bkgerr = [10.0] * n_bins
    sig = [30.0] * n_bins

    source = {
        "binning": binning,
        "bindata": {"data": data, "bkg": bkg, "bkgerr": bkgerr, "sig": sig},
    }
    return source


def hypotest_pyhf(pdf, data):
    return pyhf.infer.mle.fit(
        data, pdf, pdf.config.suggested_init(), pdf.config.suggested_bounds()
    )


def hypotest_zfit(minimizer, nll):
    with zfit.param.set_values(nll.get_params(), np.array(nll.get_params())):
        _ = minimizer.minimize(nll)


bins = [
    # 1,
    3,
    # 10,
    # 50,
    # 100,
    # 200,
    # 400,
    # 1000,
]
bin_ids = [f"{n_bins}_bins" for n_bins in bins]


@pytest.mark.benchmark(
    # group="group-name",
    # min_time=0.1,
    max_time=20,
    min_rounds=1,
    # timer=time.time,
    disable_gc=True,
    warmup=True,
    warmup_iterations=1,
)
@pytest.mark.parametrize("n_bins", bins, ids=bin_ids)
@pytest.mark.parametrize(
    "hypotest",
    ["pyhf", "zfit"],
)
@pytest.mark.parametrize(
    "eager", [False, True], ids=lambda x: "eager" if x else "graph"
)
def test_hypotest(benchmark, n_bins, hypotest, eager):
    """Benchmark the performance of pyhf.utils.hypotest() for various numbers of bins and different backends.

    Args:
        benchmark: pytest benchmark
        backend: `pyhf` tensorlib given by pytest parameterization
        n_bins: `list` of number of bins given by pytest parameterization

    Returns:
        None
    """
    source = generate_source_static(n_bins)

    signp = source["bindata"]["sig"]
    bkgnp = source["bindata"]["bkg"]
    uncnp = source["bindata"]["bkgerr"]
    datanp = source["bindata"]["data"]

    if "pyhf" in hypotest:
        hypotest = hypotest_pyhf
        if eager:
            pyhf.set_backend("numpy")
        else:
            try:
                import jax
            except ImportError:
                return
            else:
                pyhf.set_backend("jax")

        pdf = uncorrelated_background(signp, bkgnp, uncnp)
        data = datanp + pdf.config.auxdata

        # warmup
        pyhf.infer.mle.fit(
            data, pdf, pdf.config.suggested_init(), pdf.config.suggested_bounds()
        )
        benchmark(hypotest, pdf, data)
    elif hypotest == "zfit":
        with zfit.run.set_graph_mode(not eager):
            hypotest = hypotest_zfit
            obs = zfit.Space(
                "signal",
                binning=zfit.binned.RegularBinning(
                    n_bins, -0.5, n_bins + 0.5, name="signal"
                ),
            )
            zdata = zfit.data.BinnedData.from_tensor(obs, datanp)
            zmcsig = zfit.data.BinnedData.from_tensor(obs, signp)
            zmcbkg = zfit.data.BinnedData.from_tensor(obs, bkgnp)

            shapesys = {
                f"shapesys_{i}": zfit.Parameter(f"shapesys_{i}", 1, 0.1, 10)
                for i in range(n_bins)
            }
            bkgmodel = BinnedTemplatePDFV1(zmcbkg, sysshape=shapesys)
            # sigyield = zfit.Parameter('sigyield', znp.sum(zmcsig.values()))
            mu = zfit.Parameter("mu", 1, 0.1, 10)
            # sigmodeltmp = BinnedTemplatePDFV1(zmcsig)
            sigyield = zfit.ComposedParameter(
                "sigyield",
                lambda params: params["mu"] * znp.sum(zmcsig.values()),
                params={"mu": mu},
            )
            sigmodel = BinnedTemplatePDFV1(zmcsig, extended=sigyield)
            zmodel = BinnedSumPDF([sigmodel, bkgmodel])
            unc = np.array(uncnp) / np.array(bkgnp)
            constraint = zfit.constraint.GaussianConstraint(
                list(shapesys.values()), np.ones_like(unc).tolist(), unc
            )
            nll = zfit.loss.ExtendedBinnedNLL(
                zmodel,
                zdata,
                constraints=constraint,
                # options={"numhess": False}
            )

            # minimizer = zfit.minimize.Minuit(tol=1e-3, gradient=True, mode=0, verbosity=8)
            # minimizer = zfit.minimize.NLoptMMAV1(tol=1e-3, verbosity=8)
            # minimizer = zfit.minimize.NLoptLBFGSV1()
            # minimizer = zfit.minimize.ScipyLBFGSBV1()
            minimizer = zfit.minimize.ScipyTrustConstrV1()
            # minimizer = zfit.minimize.IpyoptV1()

            nll.value()
            nll.value()
            nll.gradient()
            nll.gradient()
            nll.value_gradient_hessian(params=nll.get_params())
            benchmark(hypotest, minimizer, nll)
    assert True



#  Copyright (c) 2023 zfit

import mplhep
import numpy as np
import pytest
from matplotlib import pyplot as plt

import zfit
import zfit.models.tobinned
import zfit.z.numpy as znp
from zfit import z
from zfit._data.binneddatav1 import BinnedData
from zfit._variables.axis import RegularBinning
from zfit.models.binned_functor import BinnedSumPDF
from zfit.models.template import BinnedTemplatePDFV1

mu_true = 1.2
sigma_true = 4.1
mu_true2 = 1.01
sigma_true2 = 3.5

mu_constr = [1.3, 0.02]  # mu, sigma
sigma_constr = [3.2, 0.03]

yield_true = 30000
test_values_np = np.random.normal(loc=mu_true, scale=sigma_true, size=(yield_true, 1))
test_values_np2 = np.random.normal(loc=mu_true2, scale=sigma_true2, size=yield_true)

low, high = -24.3, 28.6
obs1 = zfit.Space(
    "obs1",
    (
        np.min([test_values_np[:, 0], test_values_np2]) - 1.4,
        np.max([test_values_np[:, 0], test_values_np2]) + 2.4,
    ),
)


def create_params1(nameadd=""):
    mu1 = zfit.Parameter(
        "mu1" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 5.0, mu_true + 5.0
    )
    sigma1 = zfit.Parameter(
        "sigma1" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 4.0,
        sigma_true + 4.0,
    )
    return mu1, sigma1


def create_params2(nameadd=""):
    mu2 = zfit.Parameter(
        "mu25" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1.0, mu_true + 1.0
    )
    sigma2 = zfit.Parameter(
        "sigma25" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 2.0,
        sigma_true + 2.0,
    )
    return mu2, sigma2


def create_params3(nameadd=""):
    mu3 = zfit.Parameter(
        "mu35" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1.0, mu_true + 1.0
    )
    sigma3 = zfit.Parameter(
        "sigma35" + nameadd,
        z.to_real(sigma_true) - 0.3,
        sigma_true - 2.0,
        sigma_true + 2.0,
    )
    yield3 = zfit.Parameter("yield35" + nameadd, yield_true + 300, 0, 10000000)
    return mu3, sigma3, yield3


def create_gauss1(obs=None):
    if obs is None:
        obs = obs1
    mu, sigma = create_params1()
    return zfit.pdf.Gauss(mu, sigma, obs=obs, name="gaussian1"), mu, sigma


def create_gauss2(obs=None):
    if obs is None:
        obs = obs1
    mu, sigma = create_params2()
    return zfit.pdf.Gauss(mu, sigma, obs=obs, name="gaussian2"), mu, sigma


@pytest.mark.parametrize(
    "Loss", [zfit.loss.ExtendedBinnedNLL, zfit.loss.ExtendedBinnedChi2]
)
def test_binned_extended_simple(Loss):
    # zfit.run.set_graph_mode(False)
    counts = np.random.uniform(high=1, size=(10, 20))  # generate counts
    counts2 = np.random.normal(loc=5, size=(10, 20))
    counts3 = np.linspace(0, 10, num=10)[:, None] * np.linspace(0, 5, num=20)[None, :]
    binning = [
        zfit.binned.VariableBinning(
            [-10, -5, 5, 7, 10, 15, 17, 21, 23.5, 27, 30], name="obs1"
        ),
        RegularBinning(20, 0, 10, name="obs2"),
    ]
    obs = zfit.Space(obs=["obs1", "obs2"], binning=binning)

    mc1 = BinnedData.from_tensor(
        space=obs, values=counts, variances=znp.ones_like(counts) * 1.3
    )
    mc2 = BinnedData.from_tensor(obs, counts2)
    mc3 = BinnedData.from_tensor(obs, counts3)
    sum_counts = counts + counts2 + counts3
    observed_data = BinnedData.from_tensor(
        space=obs, values=sum_counts, variances=(sum_counts + 0.5) ** 2
    )

    pdf = BinnedTemplatePDFV1(data=mc1)
    pdf2 = BinnedTemplatePDFV1(data=mc2)
    pdf3 = BinnedTemplatePDFV1(data=mc3)
    pdf_sum = BinnedSumPDF(pdfs=[pdf, pdf2, pdf3], obs=obs)

    nll = Loss(pdf_sum, data=observed_data)
    nll2 = Loss(pdf_sum, data=observed_data)
    nll.value(), nll.gradient()  # TODO: add some check?

    nllsum = nll + nll2  # check that sum works
    # TODO: should this actually work I think?
    assert float(nllsum.value()) == pytest.approx(nll.value() + nll2.value(), rel=1e-3)


@pytest.mark.plots
@pytest.mark.flaky(2)
@pytest.mark.parametrize(
    "weights",
    [None, np.random.normal(loc=1.0, scale=0.2, size=test_values_np.shape[0])],
    ids=["no_weights", "weights_normal"],
)
@pytest.mark.parametrize(
    "Loss",
    [
        zfit.loss.BinnedNLL,
        zfit.loss.BinnedChi2,
        zfit.loss.ExtendedBinnedNLL,
        zfit.loss.ExtendedBinnedChi2,
    ],
)
@pytest.mark.parametrize(
    "simultaneous", [True, False], ids=["simultaneous", "sequential"]
)
def test_binned_loss(weights, Loss, simultaneous):
    obs = zfit.Space("obs1", limits=(-15, 25))
    gaussian1, mu1, sigma1 = create_gauss1(obs=obs)
    gaussian2, mu2, sigma2 = create_gauss2(obs=obs)
    test_values_np_shifted = test_values_np - 1.8  # shift them a bit
    test_values_np_shifted *= 1.2
    test_values = znp.array(test_values_np_shifted)
    test_values = zfit.Data.from_tensor(obs=obs, tensor=test_values, weights=weights)
    init_yield = test_values_np.shape[0] * 1.2
    scale = zfit.Parameter("yield", init_yield, 0, init_yield * 4, step_size=1)
    binning = zfit.binned.RegularBinning(92, obs.lower[0], obs.upper[0], name="obs1")
    obs_binned = obs.with_binning(binning)
    test_values_binned = test_values.to_binned(obs_binned)
    binned_gauss = zfit.pdf.BinnedFromUnbinnedPDF(gaussian1, obs_binned, extended=scale)
    binned_gauss_alt = gaussian2.to_binned(obs_binned, extended=scale)
    counts = binned_gauss.counts()
    counts_alt = binned_gauss_alt.counts()
    assert np.allclose(counts, counts_alt)
    binned_gauss_closure = binned_gauss.to_binned(obs_binned)
    assert np.allclose(counts, binned_gauss_closure.counts())
    if simultaneous:
        obs_binned2 = obs.with_binning(14)
        test_values_binned2 = test_values.to_binned(obs_binned2)
        binned_gauss2 = zfit.pdf.BinnedFromUnbinnedPDF(
            gaussian1, obs_binned2, extended=scale
        )
        loss = Loss(
            [binned_gauss, binned_gauss2],
            data=[test_values_binned, test_values_binned2],
        )

    else:
        loss = Loss(model=binned_gauss, data=test_values_binned)

    title = (
        f"Binned gaussian fit"
        f"{' (randomly weighted)' if weights is not None else ''} with "
        f"{loss.name}"
    )
    plt.figure()
    plt.title(title)
    mplhep.histplot(binned_gauss.to_hist(), label="PDF before fit")
    mplhep.histplot(test_values_binned.to_hist(), label="Data")

    # timing, uncomment to test
    # loss.value_gradient(params=loss.get_params())
    # loss.value()
    # loss.gradient()
    # import time, progressbar
    # start = time.time()
    # for _ in progressbar.progressbar(range(1000)):
    #     loss.value()
    #     loss.gradient()
    # print(f"Needed: {time.time() - start}")

    minimizer = zfit.minimize.Minuit(gradient=False)
    result = minimizer.minimize(loss=loss)

    params = result.params
    mplhep.histplot(binned_gauss.to_hist(), label="PDF after fit")
    plt.legend()
    pytest.zfit_savefig()

    result.hesse(name="hesse")
    result.errors(name="asymerr")
    str(result)  # check if no error
    rel_tol_errors = 0.1
    mu_error = 0.03 if not simultaneous else 0.021
    sigma_error = 0.0156 if simultaneous else 0.022
    params_list = [mu1, sigma1]
    errors = [mu_error, sigma_error]
    if loss.is_extended:
        params_list.append(scale)
        errors.append(122 if simultaneous else 170)
    for param, errorval in zip(params_list, errors):
        assert (
            pytest.approx(result.params[param]["hesse"]["error"], rel=rel_tol_errors)
            == errorval
        )
        assert (
            pytest.approx(result.params[param]["asymerr"]["lower"], rel=rel_tol_errors)
            == -errorval
        )
        assert (
            pytest.approx(result.params[param]["asymerr"]["upper"], rel=rel_tol_errors)
            == errorval
        )

    abs_tol_val = 0.15 if weights is None else 0.08  # more fluctuating with weights
    abs_tol_val *= 2 if isinstance(loss, zfit.loss.BinnedChi2) else 1

    assert params[mu1]["value"] == pytest.approx(
        np.mean(test_values_np_shifted), abs=abs_tol_val
    )
    assert params[sigma1]["value"] == pytest.approx(
        np.std(test_values_np_shifted), abs=abs_tol_val
    )
    if loss.is_extended:
        nexpected = test_values_np_shifted.shape[0]
        assert params[scale]["value"] == pytest.approx(
            nexpected, abs=3 * nexpected**0.5
        )
    constraints = zfit.constraint.GaussianConstraint(
        params=[mu2, sigma2],
        observation=[mu_constr[0], sigma_constr[0]],
        uncertainty=[mu_constr[1], sigma_constr[1]],
    )
    gaussian2 = zfit.models.tobinned.BinnedFromUnbinnedPDF(
        gaussian2, obs_binned, extended=scale
    )
    loss = Loss(model=gaussian2, data=test_values_binned, constraints=constraints)

    minimizer = zfit.minimize.Minuit(gradient=False)
    result = minimizer.minimize(loss=loss, params=[mu2, sigma2])
    params = result.params
    if weights is None:
        assert params[mu2]["value"] > np.mean(test_values_np_shifted)
        assert params[sigma2]["value"] < np.std(test_values_np_shifted)


@pytest.mark.parametrize("Loss", [zfit.loss.BinnedChi2, zfit.loss.ExtendedBinnedChi2])
@pytest.mark.parametrize(
    "empty", [None, "ignore", False], ids=["empty", "ignore", "False"]
)
@pytest.mark.parametrize(
    "errors",
    [None, "expected", "data"],
    ids=["error_default", "error_expected", "error_data"],
)
def test_binned_chi2_loss(Loss, empty, errors):  # TODO: add test with zeros in bins
    obs = zfit.Space("obs1", limits=(-1, 2))
    gaussian1, mu1, sigma1 = create_gauss1(obs=obs)
    test_values_np_shifted = test_values_np - 1.8  # shift them a bit
    test_values_np_shifted *= 1.2
    test_values = znp.array(test_values_np_shifted)
    test_values = zfit.Data.from_tensor(obs=obs, tensor=test_values)
    init_yield = test_values_np.shape[0] * 1.2
    scale = zfit.Parameter("yield", init_yield, 0, init_yield * 4, step_size=1)
    binning = zfit.binned.RegularBinning(32, obs.lower[0], obs.upper[0], name="obs1")
    obs_binned = obs.with_binning(binning)
    test_values_binned = test_values.to_binned(obs_binned)
    binned_gauss = zfit.models.tobinned.BinnedFromUnbinnedPDF(
        gaussian1, obs_binned, extended=scale
    )

    loss = Loss(
        model=binned_gauss,
        data=test_values_binned,
        options={"empty": empty, "errors": errors},
    )
    loss.value_gradient(loss.get_params())


@pytest.mark.parametrize(
    "weights",
    [None, np.random.normal(loc=1.0, scale=0.2, size=test_values_np.shape[0])],
    ids=["weights_none", "weights_random"],
)
@pytest.mark.parametrize(
    "Loss",
    [
        zfit.loss.BinnedNLL,
        zfit.loss.BinnedChi2,
        zfit.loss.ExtendedBinnedNLL,
        zfit.loss.ExtendedBinnedChi2,
    ],
)
def test_binned_loss_hist(weights, Loss):
    obs = zfit.Space("obs1", limits=(-15, 25))
    gaussian1, mu1, sigma1 = create_gauss1(obs=obs)
    test_values_np_shifted = test_values_np - 1.8  # shift them a bit
    test_values_np_shifted *= 1.2
    test_values = znp.array(test_values_np_shifted)
    test_values = zfit.Data.from_tensor(obs=obs, tensor=test_values, weights=weights)
    init_yield = test_values_np.shape[0] * 1.2
    scale = zfit.Parameter("yield", init_yield, 0, init_yield * 4, step_size=1)
    binning = zfit.binned.RegularBinning(32, obs.lower[0], obs.upper[0], name="obs1")
    obs_binned = obs.with_binning(binning)
    test_values_binned = test_values.to_binned(obs_binned)
    h = test_values_binned.to_hist()
    binned_gauss = zfit.models.tobinned.BinnedFromUnbinnedPDF(
        gaussian1, obs_binned, extended=scale
    )

    loss = Loss(model=binned_gauss, data=h)
    loss2 = Loss(model=binned_gauss, data=test_values_binned)

    assert pytest.approx(float(loss.value(full=True))) == float(loss2.value(full=True))

    nllsum = loss + loss2  # check that sum works
    nllsum += loss2  # check that sum works
    assert float(nllsum.value(full=True)) == pytest.approx(
        float(
            loss.value(full=True) + 2 * loss2.value(full=True)  # we add loss2 two times
        ),
        rel=1e-3,
    )



#  Copyright (c) 2023 zfit
import time

import hist
import mplhep
import numpy as np
import pytest
from matplotlib import pyplot as plt

import zfit.pdf
import zfit.z.numpy as znp
from zfit.models.interpolation import SplinePDF
from zfit.models.tobinned import BinnedFromUnbinnedPDF


@pytest.mark.plots
def test_spline_from_binned_from_unbinned():
    n = 1004
    gauss, gauss_binned, obs, obs_binned = create_gauss_binned(n)

    x = znp.linspace(-5, 10, n // 5)

    # values = gauss_binned.rel_counts(obs_binned)

    sample = gauss_binned.sample(n, limits=obs_binned)

    title = "Comparison of binned gaussian and sample"
    plt.figure()
    plt.title(title)
    mplhep.histplot(sample.to_hist(), label="sampled binned")
    plt.plot(
        obs_binned.binning.centers[0],
        gauss_binned.counts(obs_binned),
        label="counts binned",
    )
    plt.legend()
    pytest.zfit_savefig()

    spline_gauss = SplinePDF(gauss_binned, obs=obs)
    y = spline_gauss.ext_pdf(x)
    y_true = gauss.ext_pdf(x)
    plt.figure()
    plt.title("Comparison of unbinned gauss to binned to interpolated")
    plt.plot(
        obs_binned.binning.centers[0],
        gauss_binned.ext_pdf(obs_binned),
        "x",
        label="binned",
    )
    plt.plot(x, y_true, label="original")
    plt.plot(x, y, ".", label="interpolated")
    plt.legend()
    pytest.zfit_savefig()

    np.testing.assert_allclose(y, y_true, atol=50)


def test_spline2D_from_binned_from_unbinned():
    n = 1204
    gauss, gauss_binned, obs, obs_binned = create_gauss2d_binned(n, 13)

    x = znp.random.uniform([-5, 50], [10, 600], size=(1000, 2))
    data = zfit.Data.from_tensor(obs, x)

    spline_gauss = SplinePDF(gauss_binned, obs=obs)
    y = spline_gauss.ext_pdf(data)
    y_true = gauss.ext_pdf(data)

    np.testing.assert_allclose(y, y_true, atol=50)


@pytest.mark.plots
def test_unbinned_from_binned_from_unbinned():
    n = 1004
    gauss, gauss_binned, obs, obs_binned = create_gauss_binned(n)

    x = znp.linspace(-5, 10, n // 5)

    # values = gauss_binned.rel_counts(obs_binned)

    sample = gauss_binned.sample(n, limits=obs_binned)

    title = "Comparison of binned gaussian and sample"
    plt.figure()
    plt.title(title)
    mplhep.histplot(sample.to_hist(), label="sampled binned")
    plt.plot(
        obs_binned.binning.centers[0],
        gauss_binned.counts(obs_binned),
        label="counts binned",
    )
    plt.legend()
    pytest.zfit_savefig()

    unbinned = zfit.pdf.UnbinnedFromBinnedPDF(gauss_binned, obs=obs)
    unbinned2 = gauss_binned.to_unbinned()
    y = unbinned.ext_pdf(x)
    y2 = unbinned2.ext_pdf(x)
    assert np.allclose(y, y2)
    y3 = unbinned.to_unbinned().ext_pdf(x)
    assert np.allclose(y, y3)
    y_true = gauss.ext_pdf(x)
    plt.figure()
    plt.title("Comparison of unbinned gauss to binned to unbinned again")
    plt.plot(
        obs_binned.binning.centers[0],
        gauss_binned.ext_pdf(obs_binned),
        "x",
        label="binned",
    )
    plt.plot(x, y_true, label="original")
    plt.plot(x, y, ".", label="unbinned")
    plt.legend()
    pytest.zfit_savefig()
    np.testing.assert_allclose(y, y_true, atol=50)

    nsample = 500000
    sample_binned = unbinned.sample(nsample).to_binned(obs_binned)
    sample_binned_hist = sample_binned.to_hist()
    sample_gauss = gauss.sample(nsample).to_binned(obs_binned)
    sample_gauss_hist = sample_gauss.to_hist()

    title = "Comparison of unbinned gaussian and unbinned from binned sampled"
    plt.figure()
    plt.title(title)
    mplhep.histplot(sample_binned_hist, label="unbinned from binned")
    mplhep.histplot(sample_gauss_hist, label="original")
    plt.legend()
    pytest.zfit_savefig()

    diff = (sample_binned_hist.values() - sample_gauss_hist.values()) / (
        sample_gauss_hist.variances() + 1
    ) ** 0.5
    np.testing.assert_allclose(diff, 0, atol=7)  # 7 sigma away


def test_2D_unbinned_from_binned_from_unbinned():
    n = 1204
    gauss, gauss_binned, obs, obs_binned = create_gauss2d_binned(n, 13)

    x = znp.random.uniform([-5, 50], [10, 600], size=(1000, 2))
    data = zfit.Data.from_tensor(obs, x)

    unbinned = zfit.pdf.UnbinnedFromBinnedPDF(gauss_binned, obs=obs)
    y = unbinned.ext_pdf(x)
    y_true = gauss.ext_pdf(data)

    np.testing.assert_allclose(y, y_true, atol=50)

    nsample = 500000
    unbinned_sample = unbinned.sample(nsample)
    sample_binned = unbinned_sample.to_binned(obs_binned)
    sample_binned_hist = sample_binned.to_hist()
    sample_gauss = gauss.sample(nsample).to_binned(obs_binned)
    sample_gauss_hist = sample_gauss.to_hist()

    diff = (sample_binned_hist.values() - sample_gauss_hist.values()) / (
        sample_gauss_hist.variances() + 1
    ) ** 0.5
    np.testing.assert_allclose(diff, 0, atol=7)  # 7 sigma away


@pytest.mark.plots
def test_unbinned_data():
    n = 751
    gauss, gauss_binned, obs, obs_binned = create_gauss_binned(n, 70)
    x = znp.linspace(-5, 10, 200)
    centers = obs_binned.binning.centers[0]
    y_binned = gauss_binned.pdf(x)
    y_true = gauss.pdf(x)
    max_error = np.max(y_true) / 10
    np.testing.assert_allclose(y_true, y_binned, atol=max_error)

    ycenter_binned = gauss_binned.pdf(centers)
    ycenter_true = gauss.pdf(centers)
    np.testing.assert_allclose(ycenter_binned, ycenter_true, atol=max_error / 10)

    x_outside = znp.array([-7.0, 3.0, 12])
    y_outside = gauss_binned.pdf(x_outside)
    assert y_outside[0] == 0
    assert y_outside[1] > 0
    assert y_outside[2] == 0

    plt.figure()
    plt.title("Binned Gauss evaluated on unbinned edges")
    plt.plot(centers, ycenter_true, label="unbinned pdf")
    plt.plot(centers, ycenter_binned, "--", label="binned pdf")
    plt.legend()
    pytest.zfit_savefig()
    # plt.show()

    plt.figure()
    plt.title("Binned Gauss evaluated on unbinned data")
    plt.plot(x, y_true, label="unbinned pdf")
    plt.plot(x, y_binned, "--", label="binned pdf")
    plt.legend()
    pytest.zfit_savefig()


def test_unbinned_data2D():
    n = 751
    gauss, gauss_binned, obs, obs_binned = create_gauss2d_binned(n, 50)

    data = znp.random.uniform([-5, 50], [10, 600], size=(1000, 2))
    y_binned = gauss_binned.pdf(data)
    y_true = gauss.pdf(data)
    max_error = np.max(y_true) / 10
    np.testing.assert_allclose(y_true, y_binned, atol=max_error)

    centers = obs_binned.binning.centers
    X, Y = znp.meshgrid(*centers, indexing="ij")
    centers = znp.stack([znp.reshape(t, (-1,)) for t in (X, Y)], axis=-1)
    ycenter_binned = gauss_binned.pdf(centers)
    ycenter_true = gauss.pdf(centers)
    np.testing.assert_allclose(ycenter_binned, ycenter_true, atol=max_error / 10)

    # for the extended case
    y_binned_ext = gauss_binned.ext_pdf(data)
    y_true_ext = gauss.ext_pdf(data)
    max_error_ext = np.max(y_true_ext) / 10
    np.testing.assert_allclose(y_true_ext, y_binned_ext, atol=max_error_ext)

    ycenter_binned_ext = gauss_binned.ext_pdf(centers)
    ycenter_true_ext = gauss.ext_pdf(centers)
    np.testing.assert_allclose(
        ycenter_binned_ext, ycenter_true_ext, atol=max_error_ext / 10
    )

    x_outside = znp.array([[-7.0, 55], [3.0, 13], [2, 150], [12, 30], [14, 1000]])
    y_outside = gauss_binned.pdf(x_outside)
    assert y_outside[0] == 0
    assert y_outside[1] == 0
    assert y_outside[2] > 0
    assert y_outside[3] == 0
    assert y_outside[4] == 0

    y_outside_ext = gauss_binned.ext_pdf(x_outside)
    assert y_outside_ext[0] == 0
    assert y_outside_ext[1] == 0
    assert y_outside_ext[2] > 0
    assert y_outside_ext[3] == 0
    assert y_outside_ext[4] == 0


def create_gauss_binned(n, nbins=130):
    mu = zfit.Parameter("mu", 1, 0, 19)
    sigma = zfit.Parameter("sigma", 1, 0, 19)
    obs = zfit.Space("x", (-5, 10))
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    gauss.set_yield(n)
    axis = zfit.binned.RegularBinning(nbins, -5, 10, name="x")
    obs_binned = zfit.Space("x", binning=[axis])
    gauss_binned = BinnedFromUnbinnedPDF(pdf=gauss, space=obs_binned, extended=n)
    return gauss, gauss_binned, obs, obs_binned


def create_gauss2d_binned(n, nbins=130):
    mu = zfit.Parameter("mu", 1, 0, 19)
    sigma = zfit.Parameter("sigma", 1, 0, 19)
    obsx = zfit.Space("x", (-5, 10))
    obsy = zfit.Space("y", (50, 600))
    obs2d = obsx * obsy
    gaussx = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obsx)
    gaussy = zfit.pdf.Gauss(mu=250, sigma=200, obs=obsy)
    prod = zfit.pdf.ProductPDF([gaussx, gaussy])
    prod.set_yield(n)
    if isinstance(nbins, int):
        nbins = [nbins, nbins]
    axisx = zfit.binned.RegularBinning(nbins[0], -5, 10, name="x")
    axisy = zfit.binned.RegularBinning(nbins[1], 50, 600, name="y")
    obs_binned = zfit.Space(["x", "y"], binning=[axisx, axisy])
    gauss_binned = BinnedFromUnbinnedPDF(pdf=prod, space=obs_binned, extended=n)
    return prod, gauss_binned, obs2d, obs_binned


# TODO(Jonas): add test for binned pdf with unbinned data
# def test_binned_with_unbinned_data():
#     n = 100000
#
#     mu = zfit.Parameter("mu", 1, 0, 19)
#     sigma = zfit.Parameter("sigma", 1, 0, 19)
#     obs = zfit.Space("x", (-5, 10))
#     gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
#     gauss.set_yield(n)
#     axis = zfit.binned.RegularBinning(130, -5, 10, name="x")
#     obs_binned = zfit.Space("x", binning=[axis])
#     gauss_binned = BinnedFromUnbinnedPDF(pdf=gauss, space=obs_binned, extended=n)
#
#     data = znp.random.uniform(-5, 10, size=(1000,))
#     y_binned = gauss_binned.pdf(data)
#     # check shape
#     assert y_binned.shape[0] == data.shape[0]


def test_binned_from_unbinned_2D():
    n = 100000

    mu = zfit.Parameter("mu", 1, 0, 19)
    sigma = zfit.Parameter("sigma", 6, 0, 120)
    obsx = zfit.Space("x", (-5, 10))
    obsy = zfit.Space("y", (-50, 100))
    gaussx = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obsx)
    muy = mu + 3
    sigmay = sigma * 20
    gaussy = zfit.pdf.Gauss(mu=muy, sigma=sigmay, obs=obsy)
    gauss2D = zfit.pdf.ProductPDF([gaussx, gaussy])

    axisx = zfit.binned.VariableBinning(
        np.concatenate([np.linspace(-5, 5, 43), np.linspace(5, 10, 30)[1:]], axis=0),
        name="x",
    )
    axisxhist = hist.axis.Variable(
        np.concatenate([np.linspace(-5, 5, 43), np.linspace(5, 10, 30)[1:]], axis=0),
        name="x",
    )
    axisy = zfit.binned.RegularBinning(15, -50, 100, name="y")
    axisyhist = hist.axis.Regular(15, -50, 100, name="y")
    obs_binnedx = zfit.Space(["x"], binning=axisx)
    obs_binnedy = zfit.Space("y", binning=axisy)
    obs_binned = obs_binnedx * obs_binnedy

    gauss_binned = BinnedFromUnbinnedPDF(pdf=gauss2D, space=obs_binned, extended=n)
    values = gauss_binned.rel_counts(obs_binned)  # TODO: good test?
    start = time.time()
    ntrial = 10
    for _ in range(ntrial):
        values = gauss_binned.rel_counts(obs_binned)
    # print(f"Time taken {(time.time() - start) / ntrial}")
    hist2d = hist.Hist(axisxhist, axisyhist)
    nruns = 5
    npoints = 5_000_000
    for _ in range(nruns):
        normal2d = np.random.normal(
            [float(mu), float(muy)], [float(sigma), float(sigmay)], size=(npoints, 2)
        )
        hist2d.fill(*normal2d.T, threads=4)

    diff = np.abs(values * hist2d.sum() - hist2d.counts()) - 6.5 * np.sqrt(
        hist2d.counts()
    )  # 5 sigma for 1000 bins
    np.testing.assert_array_less(diff, 0)

    sample = gauss_binned.sample(n, limits=obs_binned)
    hist_sampled = sample.to_hist()
    hist_pdf = gauss_binned.to_hist()
    max_error = hist_sampled.values() * 6**2  # 6 sigma away
    np.testing.assert_array_less(
        (hist_sampled.values() - hist_pdf.values()) ** 2, max_error
    )
    plt.figure()
    plt.title("Gauss 2D binned sampled.")
    mplhep.hist2dplot(hist_sampled)
    pytest.zfit_savefig()
    plt.figure()
    plt.title("Gauss 2D binned plot, irregular (x<4.5 larger bins than x>4.5) binning.")
    mplhep.hist2dplot(hist_pdf)
    pytest.zfit_savefig()


@pytest.mark.parametrize(
    "ndim",
    [
        1,
        2,
        3,
    ],
    ids=lambda x: f"{x}D",
)
def test_binned_sampler(ndim):
    nbins = 134
    if ndim == 1:
        dims = (nbins,)
        gauss = create_gauss_binned(n=100000, nbins=dims[0])
        gauss = gauss[1]
    elif ndim == 2:
        dims = (nbins, 57)
        gauss = create_gauss2d_binned(n=100000, nbins=dims)
        obs2d = gauss[3]
        gauss = gauss[1]
    elif ndim == 3:
        dims = (nbins, 5, 3)
        obs = (
            zfit.Space("x", (-5, 10))
            * zfit.Space("y", (1, 5))
            * zfit.Space("z", (3, 6))
        )
        gauss = np.random.normal(
            loc=[0.5, 1.5, 3.6], scale=[1.2, 2.1, 0.4], size=(100000, ndim)
        )

        data = zfit.data.Data.from_numpy(obs=obs, array=gauss).to_binned(dims)
        gauss = zfit.pdf.HistogramPDF(data=data, extended=True)
    else:
        raise ValueError("ndim must be 1 or 2")
    nsampled = 100000
    sample = gauss.sample(n=nsampled)
    sampler = gauss.create_sampler(n=nsampled)

    assert sample.values().shape == dims
    assert sampler.values().shape == dims
    assert np.sum(sample.values()) == pytest.approx(nsampled)
    assert np.sum(sampler.values()) == pytest.approx(nsampled)

    sampler.resample(n=nsampled * 2)
    assert sampler.values().shape == dims
    assert np.sum(sampler.values()) == pytest.approx(nsampled * 2)

    if ndim == 2:
        sampler_swapped = sampler.with_obs(
            obs=obs2d.with_obs([obs2d.obs[1], obs2d.obs[0]])
        )
        values_swapped = sampler_swapped.values()
        assert values_swapped.shape == (dims[1], dims[0])
        assert np.sum(values_swapped) == pytest.approx(nsampled)
        sampler_swapped.resample(n=nsampled)
        assert np.sum(sampler_swapped.values()) == pytest.approx(nsampled)
        sampler_swapped.resample(n=nsampled * 3)
        assert np.sum(sampler_swapped.values()) == pytest.approx(nsampled * 3)

    sampler.resample(n=nsampled)
    # start = time.time()
    #     sampler.resample(n=nsampled)
    # print(f"Time taken {(time.time() - start)}")



#  Copyright (c) 2023 zfit

import numpy as np


def test_unbinned_histogramPDF():
    import zfit
    import zfit.z.numpy as znp

    bins1 = 5
    bins2 = 7
    counts = znp.random.uniform(high=1, size=(bins1, bins2))  # generate counts
    counts2 = np.random.normal(loc=5, size=(bins1, bins2))
    counts3 = (
        znp.linspace(0, 10, num=bins1)[:, None] * znp.linspace(0, 5, num=bins2)[None, :]
    )
    binnings = [
        zfit.binned.RegularBinning(bins1, 0, 10, name="obs1"),
        zfit.binned.RegularBinning(7, -10, bins2, name="obs2"),
    ]
    binning = binnings
    obs = zfit.Space(obs=["obs1", "obs2"], binning=binning)

    data = zfit.data.BinnedData.from_tensor(
        space=obs, values=counts, variances=znp.ones_like(counts) * 1.3
    )
    data2 = zfit.data.BinnedData.from_tensor(obs, counts2)
    data3 = zfit.data.BinnedData.from_tensor(obs, counts3)

    pdf1 = zfit.pdf.HistogramPDF(data=data, extended=znp.sum(counts))
    unbinned_pdf = zfit.pdf.UnbinnedFromBinnedPDF(pdf1)

    pdf2 = zfit.pdf.HistogramPDF(data=data2, extended=znp.sum(counts2))
    pdf3 = zfit.pdf.HistogramPDF(data=data3, extended=znp.sum(counts3))
    assert len(pdf1.ext_pdf(data)) > 0
    pdf_sum = zfit.pdf.BinnedSumPDF(pdfs=[pdf1, pdf2, pdf3], obs=obs)
    unbinned_sum = zfit.pdf.UnbinnedFromBinnedPDF(pdf_sum)
    pdfs = [unbinned_pdf, unbinned_sum]
    binned_pdfs = [pdf1, pdf_sum]

    sample = znp.random.uniform(data.space.lower, data.space.upper, size=(1000, 2))

    for pdf, binned in zip(pdfs, binned_pdfs):
        assert pdf1.is_extended
        yval = pdf.pdf(sample)
        assert yval.shape == (1000,)
        assert yval.dtype == np.float64
        assert np.min(yval) >= 0
        yval_binned = binned.pdf(sample)
        np.testing.assert_allclose(yval, yval_binned)

        integral_binned = binned.integrate(limits=obs)
        integral = pdf.integrate(limits=obs.with_binning(None))
        np.testing.assert_allclose(integral, integral_binned)
        integral_binned2 = binned.integrate(limits=obs.with_binning(None))
        np.testing.assert_allclose(integral_binned, integral_binned2)



"""Functors are functions that take typically one or more other PDF. Prominent examples are a sum, convolution etc.

A FunctorBase class is provided to make handling the models easier.

Their implementation is often non-trivial.
"""
#  Copyright (c) 2024 zfit

from __future__ import annotations

import functools
import operator
from collections import Counter
from collections.abc import Iterable
from typing import List
from typing import Optional

import pydantic
import tensorflow as tf

from typing import Literal

import zfit.z.numpy as znp
from .basefunctor import _preprocess_init_sum, FunctorPDFRepr
from .. import z
from ..core.basepdf import BasePDF
from ..core.interfaces import ZfitData, ZfitPDF
from ..core.serialmixin import SerializableMixin
from ..core.space import supports
from ..models.basefunctor import FunctorMixin, extract_daughter_input_obs
from ..serialization import Serializer
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import (
    AnalyticIntegralNotImplemented,
    NormRangeUnderdefinedError,
    SpecificFunctionNotImplemented,
)
from ..util.ztyping import ExtendedInputType, NormInputType
from ..z.random import counts_multinomial


# TODO: order of spaces if the obs is different from the wrapped pdf
class BaseFunctor(FunctorMixin, BasePDF):
    def __init__(self, pdfs, name="BaseFunctor", **kwargs):
        self.pdfs = convert_to_container(pdfs)
        super().__init__(models=self.pdfs, name=name, **kwargs)
        self._set_norm_from_daugthers()

    def _set_norm_from_daugthers(self):
        norm = super().norm
        if not norm.limits_are_set:
            norm = extract_daughter_input_obs(
                obs=norm, spaces=[model.space for model in self.models]
            )
            self.set_norm_range(norm)
        if not norm.limits_are_set:
            raise NormRangeUnderdefinedError(
                f"Daughter pdfs {self.pdfs} do not agree on a `norm` and/or no `norm`"
                "has been explicitly set."
            )

    @property
    def pdfs_extended(self):
        return [pdf.is_extended for pdf in self.pdfs]


class SumPDF(BaseFunctor, SerializableMixin):  # TODO: add extended argument
    def __init__(
        self,
        pdfs: Iterable[ZfitPDF],
        fracs: ztyping.ParamTypeInput | None = None,
        obs: ztyping.ObsTypeInput = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "SumPDF",
    ):
        """Create the sum of the `pdfs` with `fracs` as coefficients or the yields, if extended pdfs are given.

        If *all* pdfs are extended, the fracs is optional and the (normalized) yields will be used as fracs.
        If fracs is given, this will be used as the fractions, regardless of whether the pdfs have a yield or not.

        The parameters of the SumPDF are the fractions that are used to multiply the output of each daughter pdf.
        They can be accessed with `pdf.params` and have names f"frac_{i}" with i starting from 0 and going to the number
        of pdfs given.

        To get the component outputs of this pdf, e.g. to plot it, use `pdf.params.values()` to iterate through the
        fracs and `pdfs` to get the pdfs. For example

        .. code-block:: python

            for pdf, frac in zip(sumpdf.pdfs, sumpdf.params.values()):
                frac_integral = pdf.integrate(...) * frac

        Args:
            pdfs: The pdfs to be added.
            fracs: Coefficients for the linear combination of the pdfs. Optional if *all* pdfs are extended.

                - len(frac) == len(basic) - 1 results in the interpretation of a non-extended pdf.
                  The last coefficient will equal to 1 - sum(frac)
                - len(frac) == len(pdf): the fracs will be used as is and no normalization attempt is taken.
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|

        Raises
            ModelIncompatibleError: If model is incompatible.
        """
        original_init = {
            "fracs": convert_to_container(fracs),
            "extended": extended,
            "obs": obs,
        }
        # Check user input
        self._fracs = None

        pdfs = convert_to_container(pdfs)
        self.pdfs = pdfs
        (
            all_extended,
            fracs_cleaned,
            param_fracs,
            params,
            sum_yields,
            frac_param_created,
        ) = _preprocess_init_sum(fracs, obs, pdfs)
        self._frac_param_created = frac_param_created
        self._fracs = param_fracs
        self._original_fracs = fracs_cleaned
        self._automatically_extended = False

        if extended in (None, True) and all_extended and not fracs_cleaned:
            self._automatically_extended = True
            extended = sum_yields
        super().__init__(
            pdfs=pdfs, obs=obs, params=params, name=name, extended=extended, norm=norm
        )
        self.hs3.original_init.update(original_init)

    @property
    def fracs(self):
        return self._fracs

    def _apply_yield(self, value: float, norm: ztyping.LimitsType, log: bool):
        if all(self.pdfs_extended):
            return value
        else:
            return super()._apply_yield(value=value, norm=norm, log=log)

    def _unnormalized_pdf(self, x):  # NOT _pdf, as the normalization range can differ
        pdfs = self.pdfs
        fracs = self.params.values()
        probs = [pdf.pdf(x) * frac for pdf, frac in zip(pdfs, fracs)]
        prob = sum(probs)  # to keep the broadcasting ability
        return z.convert_to_tensor(prob)

    @supports(norm=True, multiple_limits=True)
    def _pdf(self, x, norm):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if norm and not equal_norm_ranges:
            raise SpecificFunctionNotImplemented
        pdfs = self.pdfs
        fracs = self.params.values()
        probs = [pdf.pdf(x) * frac for pdf, frac in zip(pdfs, fracs)]
        prob = sum(probs)
        return z.convert_to_tensor(prob)

    @supports(norm=True, multiple_limits=True)
    def _ext_pdf(self, x, norm, *, norm_range=None):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if (norm and not equal_norm_ranges) or not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        pdfs = self.pdfs
        probs = [pdf.ext_pdf(x) for pdf in pdfs]
        prob = sum(probs)
        return z.convert_to_tensor(prob)

    @supports(multiple_limits=True)
    def _integrate(self, limits, norm, options):
        pdfs = self.pdfs
        fracs = self.fracs
        # TODO(SUM): why was this needed?
        # assert norm_range not in (None, False), "Bug, who requested an unnormalized integral?"
        integrals = [
            frac
            * pdf.integrate(
                limits=limits, options=options
            )  # do NOT propagate the norm_range!
            for pdf, frac in zip(pdfs, fracs)
        ]
        return znp.sum(integrals, axis=0)

    @supports(multiple_limits=True)
    def _ext_integrate(self, limits, norm, options):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        pdfs = self.pdfs
        # TODO(SUM): why was this needed?
        # assert norm_range not in (None, False), "Bug, who requested an unnormalized integral?"
        integrals = [
            pdf.ext_integrate(
                limits=limits, options=options
            )  # do NOT propagate the norm_range!
            for pdf in pdfs
        ]
        return znp.sum(integrals, axis=0)

    @supports(multiple_limits=True)
    def _analytic_integrate(self, limits, norm):
        pdfs = self.pdfs
        fracs = self.fracs
        try:
            integrals = [
                frac
                * pdf.analytic_integrate(
                    limits=limits
                )  # do NOT propagate the norm_range!
                for pdf, frac in zip(pdfs, fracs)
            ]
        except AnalyticIntegralNotImplemented as error:
            raise AnalyticIntegralNotImplemented(
                f"analytic_integrate of pdf {self.name} is not implemented in this"
                f" SumPDF, as at least one sub-pdf does not implement it."
            ) from error

        integral = sum(integrals)
        return z.convert_to_tensor(integral)

    @supports(multiple_limits=True)
    def _partial_integrate(self, x, limits, norm, *, options):
        pdfs = self.pdfs
        fracs = self.fracs

        # do NOT propagate the norm_range!
        partial_integral = [
            pdf.partial_integrate(x=x, limits=limits, options=options) * frac
            for pdf, frac in zip(pdfs, fracs)
        ]
        partial_integral = sum(partial_integral)
        return z.convert_to_tensor(partial_integral)

    @supports(multiple_limits=True)
    def _partial_analytic_integrate(self, x, limits, norm, options):
        pdfs = self.pdfs
        fracs = self.fracs
        try:
            partial_integral = [
                pdf.partial_analytic_integrate(x=x, limits=limits) * frac
                # do NOT propagate the norm_range!
                for pdf, frac in zip(pdfs, fracs)
            ]
        except AnalyticIntegralNotImplemented as error:
            raise AnalyticIntegralNotImplemented(
                "partial_analytic_integrate of pdf {name} is not implemented in this"
                " SumPDF, as at least one sub-pdf does not implement it."
            ) from error
        partial_integral = sum(partial_integral)
        return z.convert_to_tensor(partial_integral)

    @supports(multiple_limits=True)
    def _sample(self, n, limits):
        if isinstance(n, str):
            n = [n] * len(self.pdfs)
        else:
            n = tf.unstack(counts_multinomial(total_count=n, probs=self.fracs), axis=0)

        samples = []
        for pdf, n_sample in zip(self.pdfs, n):
            sub_sample = pdf.sample(n=n_sample, limits=limits)
            if isinstance(sub_sample, ZfitData):
                sub_sample = sub_sample.value()
            samples.append(sub_sample)
        sample = znp.concatenate(samples, axis=0)
        sample = z.random.shuffle(sample)
        return sample


class SumPDFRepr(FunctorPDFRepr):
    _implementation = SumPDF
    hs3_type: Literal["SumPDF"] = pydantic.Field("SumPDF", alias="type")
    fracs: Optional[List[Serializer.types.ParamInputTypeDiscriminated]] = None

    @pydantic.root_validator(pre=True)
    def validate_all_sumpdf(cls, values):
        # if cls.orm_mode(values):
        #     init = values["hs3"].original_init
        #     values = dict(values)
        #     values["fracs"] = init["fracs"]
        return values


class ProductPDF(BaseFunctor, SerializableMixin):
    def __init__(
        self,
        pdfs: list[ZfitPDF],
        obs: ztyping.ObsTypeInput = None,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name="ProductPDF",
    ):
        """Product of multiple PDFs in the same or different variables.

        This implementation optimizes the integration: if the PDFs are in exclusive observables, the
        integration can be factorized and the integrals can be calculated separately. This also takes into account
        that only some PDFs might have overlapping observables.

        Args:
            pdfs: List of PDFs to multiply.
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
        """
        original_init = {"extended": extended, "obs": obs}

        super().__init__(pdfs=pdfs, obs=obs, name=name, extended=extended, norm=norm)
        self.hs3.original_init.update(original_init)

        same_obs_pdfs = []
        disjoint_obs_pdfs = []
        same_obs = Counter([ob for pdf in self.pdfs for ob in pdf.obs])
        same_obs = {k for k, v in same_obs.items() if v > 1}
        for pdf in self.pdfs:
            if set(pdf.obs).isdisjoint(same_obs):
                disjoint_obs_pdfs.append(pdf)
            else:
                same_obs_pdfs.append(pdf)
        if same_obs_pdfs and disjoint_obs_pdfs:
            same_obs_pdf = type(self)(pdfs=same_obs_pdfs)
            disjoint_obs_pdfs.append(same_obs_pdf)
            self.add_cache_deps(same_obs_pdfs)
            self._prod_is_same_obs_pdf = False
            self._prod_disjoint_obs_pdfs = disjoint_obs_pdfs
        elif disjoint_obs_pdfs:
            self._prod_is_same_obs_pdf = False
            self._prod_disjoint_obs_pdfs = disjoint_obs_pdfs
            assert set(disjoint_obs_pdfs) == set(self.pdfs)
        else:
            self._prod_is_same_obs_pdf = True
            self._prod_disjoint_obs_pdfs = None

    def _unnormalized_pdf(self, x: ztyping.XType):
        probs = [pdf.pdf(x, norm=False) for pdf in self.pdfs]
        prob = functools.reduce(operator.mul, probs)
        return z.convert_to_tensor(prob)

    @supports(norm=True, multiple_limits=True)
    def _pdf(self, x, norm):
        equal_norm_ranges = (
            len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        )  # all equal
        if not self._prod_is_same_obs_pdf and equal_norm_ranges:
            probs = [pdf.pdf(var=x, norm=norm) for pdf in self._prod_disjoint_obs_pdfs]
            prob = functools.reduce(operator.mul, probs)
            return z.convert_to_tensor(prob)
        else:
            raise SpecificFunctionNotImplemented

    @supports(norm=False)
    def _integrate(self, limits, norm, options):
        if not self._prod_is_same_obs_pdf:
            integrals = []
            for pdf in self._prod_disjoint_obs_pdfs:
                limit = limits.with_obs(pdf.obs)
                integrals.append(
                    pdf.integrate(limits=limit, norm=norm, options=options)
                )
            integral = functools.reduce(operator.mul, integrals)
            return z.convert_to_tensor(integral)
        else:
            raise SpecificFunctionNotImplemented

    @supports(norm=False)
    def _analytic_integrate(self, limits, norm):
        if self._prod_is_same_obs_pdf:
            raise AnalyticIntegralNotImplemented(
                f"Cannot integrate analytically as PDFs have overlapping obs:"
                f" {[pdf.obs for pdf in self.pdfs]}"
            )
        integrals = []
        for pdf in self._prod_disjoint_obs_pdfs:
            limit = limits.with_obs(pdf.obs)
            try:
                integral = pdf.analytic_integrate(limits=limit, norm=norm)
            except AnalyticIntegralNotImplemented:
                raise AnalyticIntegralNotImplemented(
                    f"At least one pdf ({pdf} does not support analytic integration."
                )
            else:
                integrals.append(integral)
        integral = functools.reduce(operator.mul, integrals)
        return z.convert_to_tensor(integral)

    @supports(multiple_limits=True, norm=True)
    def _partial_integrate(self, x, limits, norm, *, options):
        if self._prod_is_same_obs_pdf:
            raise SpecificFunctionNotImplemented
        pdfs = self._prod_disjoint_obs_pdfs

        values = []
        for pdf in pdfs:
            intersection_limits = set(pdf.obs).intersection(limits.obs)
            intersection_data = set(pdf.obs).intersection(x.obs)
            if intersection_limits and not intersection_data:
                values.append(pdf.integrate(limits=limits, norm=norm, options=options))
            elif intersection_limits:  # implicitly "and intersection_data"
                values.append(
                    pdf.partial_integrate(x=x, limits=limits, options=options)
                )
            else:
                assert (
                    not intersection_limits and intersection_data
                ), "Something slipped, the logic is flawed."
                values.append(pdf.pdf(x, norm_range=norm))
        values = functools.reduce(operator.mul, values)
        return z.convert_to_tensor(values)


class ProductPDFRepr(FunctorPDFRepr):
    _implementation = ProductPDF
    hs3_type: Literal["ProductPDF"] = pydantic.Field("ProductPDF", alias="type")



#  Copyright (c) 2022 zfit

import numpy as np
import tensorflow as tf

from zfit import z


def numpy_func(x, a):
    return np.square(x) * a


@z.function
def wrapped_numpy_func(x_tensor, a_tensor):
    result = tf.numpy_function(
        func=numpy_func, inp=[x_tensor, a_tensor], Tout=tf.float64
    )
    result = tf.sqrt(result)
    return result


def test_wrapped_func():
    rnd = z.random.uniform(shape=(10,))
    result = wrapped_numpy_func(rnd, z.constant(3.0))
    np.testing.assert_allclose(rnd * np.sqrt(3), result)


def test_multinomial():
    import zfit.z.numpy as znp

    probs = np.array([0.5, 0.5])
    x = z.random.counts_multinomial(10000, probs=probs)
    assert x.shape == (2,)
    probs = np.array([0.1, 0.2, 0.7])
    x = z.random.counts_multinomial(10000, probs=probs)
    assert x.shape == (3,)
    probs = np.array([0.5, 0.5])
    x = z.random.counts_multinomial(10000, probs=probs, dtype=znp.float64)
    assert x.shape == (2,)
    probs = np.array([0.1, 0.7, 0.2])
    x = z.random.counts_multinomial(10000, probs=probs, dtype=znp.float64)
    assert x.shape == (3,)
